from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import get_data_paths, get_is_remote_path

from fairspec_dataset.actions.file.copy import copy_file
from fairspec_dataset.actions.file.temp import get_temp_file_path

if TYPE_CHECKING:
    from fairspec_metadata import Resource


def prefetch_files(
    resource: Resource,
    *,
    max_bytes: int | None = None,
) -> list[str]:
    paths = get_data_paths(resource)
    if not paths:
        return []
    return [prefetch_file(path, max_bytes=max_bytes) for path in paths]


def prefetch_file(
    path: str,
    *,
    max_bytes: int | None = None,
) -> str:
    if not get_is_remote_path(path):
        return path

    new_path = get_temp_file_path()
    copy_file(source_path=path, target_path=new_path, max_bytes=max_bytes)
    return new_path



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class GithubFile(FairspecModel):
    path: str | None = None
    mode: str | None = None
    type: str | None = None
    size: int | None = None
    sha: str | None = None
    url: str | None = None



from .actions.dataset.load import load_dataset_from_folder
from .actions.dataset.save import save_dataset_to_folder
from .plugin import FolderPlugin

__all__ = [
    "FolderPlugin",
    "load_dataset_from_folder",
    "save_dataset_to_folder",
]



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from fairspec_metadata import get_file_name_slug
from fairspec_metadata import Integrity, IntegrityType
from fairspec_metadata import Resource

if TYPE_CHECKING:
    from fairspec_dataset.plugins.zenodo.models.file import ZenodoFile


def convert_resource_from_zenodo(zenodo_file: ZenodoFile) -> Resource:
    links = zenodo_file.links
    path = _convert_path(links.self or "" if links else "")

    return Resource(
        data=path,
        name=get_file_name_slug(zenodo_file.key or "") or zenodo_file.id or "",
        integrity=Integrity(
            type=IntegrityType.md5,
            hash=(zenodo_file.checksum or "").replace("md5:", ""),
        ),
        unstable_customMetadata={
            "zenodoKey": zenodo_file.key,
            "zenodoUrl": path,
        },
    )


def _convert_path(link: str) -> str:
    result = link.replace("/api/", "/")
    result = re.sub(r"/content$", "", result)
    return result



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from fairspec_dataset.actions.dataset.merge import merge_datasets
from fairspec_dataset.plugins.zenodo.models.record import ZenodoRecord

from fairspec_dataset.plugins.zenodo.services.zenodo import make_zenodo_api_request
from .from_zenodo import convert_dataset_from_zenodo

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


def load_dataset_from_zenodo(
    dataset_url: str,
    *,
    api_key: str | None = None,
) -> Descriptor:
    parsed = urllib.parse.urlparse(dataset_url)
    sandbox = parsed.hostname == "sandbox.zenodo.org"

    record_id = _extract_record_id(dataset_url)
    if not record_id:
        raise Exception(f"Failed to extract record ID from URL: {dataset_url}")

    zenodo_record = make_zenodo_api_request(
        endpoint=f"/records/{record_id}",
        api_key=api_key,
        sandbox=sandbox,
    )

    system_dataset = convert_dataset_from_zenodo(ZenodoRecord(**zenodo_record))
    user_dataset_path: str | None = None
    for resource in system_dataset.resources or []:
        custom = resource.unstable_customMetadata or {}
        if custom.get("zenodoKey") == "dataset.json":
            user_dataset_path = custom.get("zenodoUrl")
            break

    dataset = merge_datasets(
        system_dataset=system_dataset,
        user_dataset_path=user_dataset_path,
    )

    for resource in dataset.resources or []:
        resource.unstable_customMetadata = None

    return dataset.model_dump(by_alias=True, exclude_none=True)


def _extract_record_id(dataset_url: str) -> str | None:
    parsed = urllib.parse.urlparse(dataset_url)
    parts = [p for p in parsed.path.split("/") if p]
    return parts[-1] if parts else None



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel

from .creator import ZenodoCreator


class ZenodoMetadata(FairspecModel):
    title: str | None = None
    description: str | None = None
    upload_type: str | None = None
    publication_date: str | None = None
    creators: list[ZenodoCreator] | None = None
    access_right: str | None = None
    license: str | None = None
    doi: str | None = None
    keywords: list[str] | None = None
    related_identifiers: list[dict] | None = None
    communities: list[dict] | None = None
    version: str | None = None



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class CkanTag(FairspecModel):
    id: str | None = None
    name: str | None = None
    display_name: str | None = None



from .actions.dataset.load import load_dataset_from_github
from .actions.dataset.save import save_dataset_to_github
from .plugin import GithubPlugin

__all__ = [
    "GithubPlugin",
    "load_dataset_from_github",
    "save_dataset_to_github",
]



from __future__ import annotations

import base64
from collections.abc import Callable
from typing import TYPE_CHECKING

from fairspec_metadata import denormalize_dataset, stringify_descriptor

from fairspec_dataset.actions.dataset.basepath import get_dataset_basepath
from fairspec_dataset.actions.resource.save import SaveFileProps, save_resource_files
from fairspec_dataset.actions.stream.load import load_file_stream

from fairspec_dataset.plugins.github.services.github import make_github_api_request

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor
    from fairspec_metadata import Resource


def save_dataset_to_github(
    dataset: Dataset,
    *,
    api_key: str,
    repo: str,
    org: str | None = None,
) -> dict:
    basepath = get_dataset_basepath(dataset)

    endpoint = f"/orgs/{org}/repos" if org else "/user/repos"
    github_repository = make_github_api_request(
        endpoint=endpoint,
        payload={"name": repo, "auto_init": True},
        method="POST",
        api_key=api_key,
    )

    owner_login = github_repository["owner"]["login"]

    resource_descriptors: list[Descriptor] = []
    for resource in dataset.resources or []:

        def _make_save_file(res: Resource) -> Callable[[SaveFileProps], str]:
            def _save_file(props: SaveFileProps) -> str:
                stream = load_file_stream(props.normalized_path)
                content = base64.b64encode(stream.read()).decode()

                payload = {
                    "path": props.denormalized_path,
                    "content": content,
                    "message": f'Added file "{props.denormalized_path}"',
                }

                make_github_api_request(
                    endpoint=f"/repos/{owner_login}/{repo}/contents/{props.denormalized_path}",
                    method="PUT",
                    payload=payload,
                    api_key=api_key,
                )

                return props.denormalized_path

            return _save_file

        resource_descriptors.append(
            save_resource_files(
                resource,
                basepath=basepath,
                with_remote=False,
                save_file=_make_save_file(resource),
            )
        )

    denormalized = denormalize_dataset(dataset, basepath=basepath)
    descriptor: Descriptor = {
        **denormalized.model_dump(by_alias=True, exclude_none=True),
        "resources": resource_descriptors,
    }

    content = base64.b64encode(stringify_descriptor(descriptor).encode()).decode()

    make_github_api_request(
        endpoint=f"/repos/{owner_login}/{repo}/contents/dataset.json",
        method="PUT",
        payload={
            "path": "dataset.json",
            "message": 'Added file "dataset.json"',
            "content": content,
        },
        api_key=api_key,
    )

    return {
        "path": f"https://raw.githubusercontent.com/{owner_login}/{repo}/refs/heads/main/dataset.json",
        "repo_url": github_repository["html_url"],
    }



from __future__ import annotations

from fairspec_metadata import Resource

from fairspec_dataset.models.file import FileDescription

from .infer import infer_bytes, infer_integrity, infer_textual
from .prefetch import prefetch_file


def describe_file(
    path: str,
    *,
    hash_type: str = "sha256",
) -> FileDescription:
    local_path = prefetch_file(path)
    resource = Resource(data=local_path)

    return FileDescription(
        bytes=infer_bytes(resource),
        textual=infer_textual(resource),
        integrity=infer_integrity(resource, hash_type=hash_type),
    )



from __future__ import annotations

import json
import os
import re

from fairspec_dataset.plugins.ckan.models.dataset import CkanDataset
from .from_ckan import convert_dataset_from_ckan


def _load_fixture() -> CkanDataset:
    path = os.path.join(os.path.dirname(__file__), "fixtures", "ckan-dataset.json")
    with open(path) as f:
        return CkanDataset(**json.load(f))


class TestConvertDatasetFromCkan:
    def test_converts_ckan_dataset_to_fairspec_dataset(self):
        ckan_dataset = _load_fixture()

        result = convert_dataset_from_ckan(ckan_dataset)

        assert result.titles is not None
        assert len(result.titles) == 1
        assert result.titles[0].title == ckan_dataset.title

        assert result.descriptions is not None
        assert len(result.descriptions) == 1
        assert result.descriptions[0].description == ckan_dataset.notes
        assert result.descriptions[0].descriptionType == "Abstract"

        assert result.version == ckan_dataset.version

        assert result.dates is not None
        assert len(result.dates) == 2
        created_date = next(d for d in result.dates if d.dateType == "Created")
        assert created_date.date == ckan_dataset.metadata_created
        updated_date = next(d for d in result.dates if d.dateType == "Updated")
        assert updated_date.date == ckan_dataset.metadata_modified

        assert result.rightsList is not None
        assert len(result.rightsList) == 1
        rights = result.rightsList[0]
        assert rights.rights == ckan_dataset.license_title
        assert rights.rightsUri == ckan_dataset.license_url
        assert rights.rightsIdentifier == ckan_dataset.license_id

        assert result.creators is not None
        assert len(result.creators) == 1
        assert result.creators[0].name == ckan_dataset.author
        assert result.creators[0].nameType == "Personal"

        assert result.contributors is not None
        assert len(result.contributors) == 1
        assert result.contributors[0].name == ckan_dataset.maintainer
        assert result.contributors[0].nameType == "Personal"
        assert result.contributors[0].contributorType == "ContactPerson"

        assert result.subjects is not None
        assert ckan_dataset.tags is not None
        assert len(result.subjects) == len(ckan_dataset.tags)
        assert [s.subject for s in result.subjects] == [
            tag.name for tag in ckan_dataset.tags
        ]

        assert result.resources is not None
        assert ckan_dataset.resources is not None
        assert len(result.resources) == len(ckan_dataset.resources)

        first_ckan_resource = ckan_dataset.resources[0]
        first_resource = result.resources[0]
        assert first_resource.data == first_ckan_resource.url
        assert first_resource.name is not None
        assert re.match(r"^sample[-_]linked[-_]csv$", first_resource.name)
        assert first_resource.descriptions is not None
        assert (
            first_resource.descriptions[0].description
            == first_ckan_resource.description
        )

    def test_handles_empty_resources_array(self):
        ckan_dataset = _load_fixture().model_copy(update={"resources": []})

        result = convert_dataset_from_ckan(ckan_dataset)

        assert result.resources == []

    def test_handles_undefined_optional_properties(self):
        ckan_dataset = CkanDataset(resources=[], tags=[], id="test", name="test")

        result = convert_dataset_from_ckan(ckan_dataset)

        assert result.titles is None
        assert result.descriptions is None
        assert result.version is None
        assert result.dates is None
        assert result.rightsList is None
        assert result.creators is None
        assert result.contributors is None
        assert result.subjects is None
        assert result.resources == []



from __future__ import annotations

from typing import Required, TypedDict

from fairspec_metadata.models.base import FairspecModel


class SaveDatasetOptions(TypedDict, total=False):
    target: Required[str]
    with_remote: bool


class SaveDatasetResult(FairspecModel):
    path: str | None = None



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Callable

from fairspec_metadata import (
    denormalize_path,
    get_file_name,
    get_is_remote_path,
)

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor
    from fairspec_metadata import Resource


class SaveFileProps:
    def __init__(
        self,
        *,
        property_name: str,
        property_index: int,
        normalized_path: str,
        denormalized_path: str,
    ) -> None:
        self.property_name = property_name
        self.property_index = property_index
        self.normalized_path = normalized_path
        self.denormalized_path = denormalized_path


SaveFileCallback = Callable[[SaveFileProps], str]


def save_resource_files(
    resource: Resource,
    *,
    save_file: SaveFileCallback,
    basepath: str | None = None,
    with_remote: bool = False,
    without_folders: bool = False,
) -> Descriptor:
    resource = resource.model_copy(deep=True)
    dedup_indexes: dict[str, int] = {}

    def _save_file(path: str, name: str, index: int) -> str:
        is_remote = get_is_remote_path(path)

        denormalized_path = denormalize_path(path, basepath=basepath)
        normalized_path = path

        if is_remote:
            if not with_remote:
                return path
            filename = get_file_name(path)
            if not filename:
                return path
            denormalized_path = filename
        elif without_folders:
            denormalized_path = denormalized_path.replace("/", "-")

        dedup_index = dedup_indexes.get(denormalized_path, 0)
        dedup_indexes[denormalized_path] = dedup_index + 1

        if dedup_index:
            denormalized_path = re.sub(
                r"^(.*?)([^/]+?)(\.[^/]+(?:\.[^/]+)*)$",
                rf"\1\2-{dedup_index}\3",
                denormalized_path,
            )

        return save_file(
            SaveFileProps(
                property_name=name,
                property_index=index,
                normalized_path=normalized_path,
                denormalized_path=denormalized_path,
            )
        )

    if isinstance(resource.data, str):
        resource.data = _save_file(resource.data, "data", 0)

    if isinstance(resource.data, list):
        for i, item in enumerate(resource.data):
            if isinstance(item, str):
                resource.data[i] = _save_file(item, "data", i)

    for name in ("dataSchema", "tableSchema"):
        prop = getattr(resource, name, None)
        if isinstance(prop, str):
            setattr(resource, name, _save_file(prop, name, 0))

    return resource.model_dump(by_alias=True, exclude_none=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Unpack

from fairspec_metadata.plugin import MetadataPlugin

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor
    from fairspec_metadata import FileDialect
    from fairspec_metadata import Resource

    from .models.dataset import SaveDatasetOptions, SaveDatasetResult
    from .models.file_dialect import InferFileDialectOptions


class DatasetPlugin(MetadataPlugin):
    def load_dataset(self, source: str) -> Descriptor | None:
        return None

    def save_dataset(
        self, dataset: Dataset, **options: Unpack[SaveDatasetOptions]
    ) -> SaveDatasetResult | None:
        return None

    def infer_file_dialect(
        self,
        resource: Resource,
        **options: Unpack[InferFileDialectOptions],
    ) -> FileDialect | None:
        return None



from __future__ import annotations

from fairspec_dataset.actions.stream.load import load_file_stream
from fairspec_dataset.actions.stream.save import save_file_stream


def copy_file(
    *,
    source_path: str,
    target_path: str,
    max_bytes: int | None = None,
) -> None:
    stream = load_file_stream(source_path, max_bytes=max_bytes)
    save_file_stream(stream, path=target_path)



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel

from .organization import CkanOrganization
from .resource import CkanResource
from .tag import CkanTag


class CkanDataset(FairspecModel):
    resources: list[CkanResource] | None = None
    organization: CkanOrganization | None = None
    tags: list[CkanTag] | None = None
    id: str | None = None
    name: str | None = None
    title: str | None = None
    notes: str | None = None
    version: str | None = None
    license_id: str | None = None
    license_title: str | None = None
    license_url: str | None = None
    author: str | None = None
    author_email: str | None = None
    maintainer: str | None = None
    maintainer_email: str | None = None
    metadata_created: str | None = None
    metadata_modified: str | None = None



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class GithubLicense(FairspecModel):
    key: str | None = None
    name: str | None = None
    spdx_id: str | None = None
    url: str | None = None



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel

from fairspec_metadata import Integrity


class FileDescription(FairspecModel):
    bytes: int
    textual: bool
    integrity: Integrity | None



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class CkanFieldInfo(FairspecModel):
    label: str | None = None
    notes: str | None = None
    type_override: str | None = None


class CkanField(FairspecModel):
    id: str | None = None
    type: str | None = None
    info: CkanFieldInfo | None = None



from unittest.mock import MagicMock, patch

from fairspec_metadata import Dataset
from fairspec_metadata import Resource

from .plugin import DescriptorPlugin


class TestLoadDataset:
    def setup_method(self):
        self.plugin = DescriptorPlugin()

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_loads_from_local_datapackage_json(self, mock_load: MagicMock):
        mock_dataset = MagicMock()
        mock_dataset.model_dump.return_value = {
            "resources": [{"name": "test", "data": []}]
        }
        mock_load.return_value = mock_dataset

        result = self.plugin.load_dataset("./datapackage.json")

        mock_load.assert_called_once_with("./datapackage.json")
        assert result == {"resources": [{"name": "test", "data": []}]}

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_loads_from_local_json(self, mock_load: MagicMock):
        mock_dataset = MagicMock()
        mock_dataset.model_dump.return_value = {
            "resources": [{"name": "test", "data": []}]
        }
        mock_load.return_value = mock_dataset

        result = self.plugin.load_dataset("./dataset.json")

        mock_load.assert_called_once_with("./dataset.json")
        assert result == {"resources": [{"name": "test", "data": []}]}

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_loads_from_absolute_path(self, mock_load: MagicMock):
        mock_dataset = MagicMock()
        mock_dataset.model_dump.return_value = {
            "resources": [{"name": "test", "data": []}]
        }
        mock_load.return_value = mock_dataset

        result = self.plugin.load_dataset("/absolute/path/datapackage.json")

        mock_load.assert_called_once_with("/absolute/path/datapackage.json")
        assert result == {"resources": [{"name": "test", "data": []}]}

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_returns_none_for_csv(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("./data.csv")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_returns_none_for_xlsx(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("./data.xlsx")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_returns_none_for_parquet(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("./data.parquet")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.load_dataset_descriptor")
    def test_returns_none_for_zenodo_url(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://zenodo.org/record/123")

        mock_load.assert_not_called()
        assert result is None


class TestSaveDataset:
    def setup_method(self):
        self.plugin = DescriptorPlugin()
        self.dataset = Dataset(resources=[Resource(name="test", data=[])])

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_saves_to_local_datapackage_json(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="./datapackage.json")

        mock_save.assert_called_once_with(self.dataset, path="./datapackage.json")
        assert result is not None
        assert result.path == "./datapackage.json"

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_saves_with_absolute_path(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="/absolute/path/datapackage.json"
        )

        mock_save.assert_called_once_with(
            self.dataset, path="/absolute/path/datapackage.json"
        )
        assert result is not None
        assert result.path == "/absolute/path/datapackage.json"

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_returns_none_for_remote_https(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="https://example.com/datapackage.json"
        )

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_returns_none_for_remote_http(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="http://example.com/datapackage.json"
        )

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_returns_none_for_non_datapackage_json(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="./dataset.json")

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_returns_none_for_csv(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="./data.csv")

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_returns_none_for_xlsx(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="./data.xlsx")

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_returns_none_for_directory(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="./data")

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.descriptor.plugin.save_dataset_descriptor")
    def test_ignores_with_remote_option(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="./datapackage.json", with_remote=True
        )

        mock_save.assert_called_once_with(self.dataset, path="./datapackage.json")
        assert result is not None
        assert result.path == "./datapackage.json"



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class GithubOwner(FairspecModel):
    login: str | None = None
    id: int | None = None
    avatar_url: str | None = None
    html_url: str | None = None
    type: str | None = None



from __future__ import annotations

import os
import shutil
from typing import BinaryIO


def save_file_stream(
    stream: BinaryIO,
    *,
    path: str,
    overwrite: bool = False,
) -> None:
    os.makedirs(os.path.dirname(path) or ".", exist_ok=True)

    if not overwrite and os.path.exists(path):
        raise FileExistsError(f'Path "{path}" already exists')

    with open(path, "wb") as f:
        shutil.copyfileobj(stream, f)



from __future__ import annotations

import atexit
import shutil
import tempfile


def get_temp_folder_path(*, persist: bool = False) -> str:
    path = tempfile.mkdtemp()

    if not persist:
        atexit.register(_cleanup_dir, path)

    return path


def _cleanup_dir(path: str) -> None:
    try:
        shutil.rmtree(path)
    except OSError:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import (
    ContributorType,
    CreatorNameType,
    DateType,
    DescriptionType,
)
from fairspec_metadata import Contributor
from fairspec_metadata import Creator
from fairspec_metadata import DataciteDate
from fairspec_metadata import DataciteDescription
from fairspec_metadata import Rights
from fairspec_metadata import Subject
from fairspec_metadata import Title
from fairspec_metadata import Dataset

from fairspec_dataset.plugins.github.actions.resource.from_github import (
    convert_resource_from_github,
)

if TYPE_CHECKING:
    from fairspec_dataset.plugins.github.models.repository import GithubRepository


def convert_dataset_from_github(repository: GithubRepository) -> Dataset:
    titles = [Title(title=repository.full_name)] if repository.full_name else None

    descriptions = (
        [
            DataciteDescription(
                description=repository.description,
                descriptionType=DescriptionType.Abstract,
            )
        ]
        if repository.description
        else None
    )

    rights_list = None
    license_info = repository.license
    if license_info:
        rights_list = [
            Rights(
                rights=license_info.name or "",
                rightsUri=license_info.url,
                rightsIdentifier=license_info.spdx_id or license_info.key,
                rightsIdentifierScheme="SPDX",
            )
        ]

    creators = None
    contributors = None
    owner = repository.owner
    if owner:
        if owner.type == "Organization":
            contributors = [
                Contributor(
                    name=owner.login or "",
                    nameType=CreatorNameType.Organizational,
                    contributorType=ContributorType.HostingInstitution,
                )
            ]
        else:
            creators = [
                Creator(
                    name=owner.login or "",
                    nameType=CreatorNameType.Personal,
                )
            ]

    files = repository.files or []
    resource_list = []
    if files:
        default_branch = repository.default_branch or "main"
        resource_list = [
            convert_resource_from_github(f, default_branch=default_branch)
            for f in files
            if not (f.path or "").startswith(".") and f.type == "blob"
        ]

    topics = repository.topics or []
    subjects = [Subject(subject=topic) for topic in topics] if topics else None

    dates = (
        [DataciteDate(date=repository.created_at, dateType=DateType.Created)]
        if repository.created_at
        else None
    )

    return Dataset(
        titles=titles,
        descriptions=descriptions,
        rightsList=rights_list,
        creators=creators,
        contributors=contributors,
        resources=resource_list,
        subjects=subjects,
        dates=dates,
    )



from __future__ import annotations

import json
import urllib.request
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


def make_github_api_request(
    *,
    endpoint: str,
    method: str = "GET",
    payload: Descriptor | None = None,
    api_key: str | None = None,
) -> dict:
    base_url = "https://api.github.com"
    url = f"{base_url}{endpoint}"

    headers: dict[str, str] = {}
    if api_key:
        headers["Authorization"] = f"Bearer {api_key}"

    data: bytes | None = None
    if payload:
        headers["Content-Type"] = "application/json"
        data = json.dumps(payload).encode()

    request = urllib.request.Request(url, data=data, headers=headers, method=method)

    with urllib.request.urlopen(request) as response:
        return json.loads(response.read().decode())



from unittest.mock import MagicMock, patch

from fairspec_metadata import Dataset
from fairspec_metadata import Resource

from .plugin import FolderPlugin


class TestLoadDataset:
    def setup_method(self):
        self.plugin = FolderPlugin()

    @patch("fairspec_dataset.plugins.folder.plugin.os.path.isdir", return_value=True)
    @patch("fairspec_dataset.plugins.folder.plugin.load_dataset_from_folder")
    def test_loads_from_local_directory(
        self, mock_load: MagicMock, _mock_isdir: MagicMock
    ):
        mock_dataset = MagicMock()
        mock_dataset.model_dump.return_value = {
            "resources": [{"name": "test", "data": []}]
        }
        mock_load.return_value = mock_dataset

        result = self.plugin.load_dataset(".")

        mock_load.assert_called_once_with(".")
        assert result == {"resources": [{"name": "test", "data": []}]}

    @patch("fairspec_dataset.plugins.folder.plugin.load_dataset_from_folder")
    def test_returns_none_for_http(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("http://example.com/data")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.folder.plugin.load_dataset_from_folder")
    def test_returns_none_for_https(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://example.com/data")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.folder.plugin.load_dataset_from_folder")
    def test_returns_none_for_ftp(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("ftp://example.com/data")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.folder.plugin.load_dataset_from_folder")
    def test_returns_none_for_github(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://github.com/owner/repo/data")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.folder.plugin.load_dataset_from_folder")
    def test_returns_none_for_zenodo(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://zenodo.org/record/123")

        mock_load.assert_not_called()
        assert result is None


class TestSaveDataset:
    def setup_method(self):
        self.plugin = FolderPlugin()
        self.dataset = Dataset(resources=[Resource(name="test", data=[])])

    @patch("fairspec_dataset.plugins.folder.plugin.os.path.isdir", return_value=True)
    @patch("fairspec_dataset.plugins.folder.plugin.save_dataset_to_folder")
    def test_saves_to_local_directory(
        self, mock_save: MagicMock, _mock_isdir: MagicMock
    ):
        result = self.plugin.save_dataset(self.dataset, target="/tmp/test")

        mock_save.assert_called_once_with(
            self.dataset, folder_path="/tmp/test", with_remote=False
        )
        assert result is not None
        assert result.path == "/tmp/test"

    @patch("fairspec_dataset.plugins.folder.plugin.os.path.isdir", return_value=True)
    @patch("fairspec_dataset.plugins.folder.plugin.save_dataset_to_folder")
    def test_saves_with_remote_option(
        self, mock_save: MagicMock, _mock_isdir: MagicMock
    ):
        result = self.plugin.save_dataset(
            self.dataset, target="/tmp/test", with_remote=True
        )

        mock_save.assert_called_once_with(
            self.dataset, folder_path="/tmp/test", with_remote=True
        )
        assert result is not None
        assert result.path == "/tmp/test"

    @patch("fairspec_dataset.plugins.folder.plugin.save_dataset_to_folder")
    def test_returns_none_for_http(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="http://example.com/data"
        )

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.folder.plugin.save_dataset_to_folder")
    def test_returns_none_for_https(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="https://example.com/data"
        )

        mock_save.assert_not_called()
        assert result is None



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from fairspec_metadata import denormalize_dataset, stringify_descriptor

from fairspec_dataset.actions.dataset.basepath import get_dataset_basepath
from fairspec_dataset.actions.resource.save import (
    SaveFileCallback,
    SaveFileProps,
    save_resource_files,
)
from fairspec_dataset.actions.stream.load import load_file_stream

from fairspec_dataset.plugins.zenodo.services.zenodo import make_zenodo_api_request
from .to_zenodo import convert_dataset_to_zenodo

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor
    from fairspec_metadata import Resource


def save_dataset_to_zenodo(
    dataset: Dataset,
    *,
    api_key: str,
    sandbox: bool = False,
) -> dict:
    basepath = get_dataset_basepath(dataset)

    new_zenodo_record = convert_dataset_to_zenodo(dataset)
    zenodo_record = make_zenodo_api_request(
        payload=new_zenodo_record,
        endpoint="/deposit/depositions",
        method="POST",
        api_key=api_key,
        sandbox=sandbox,
    )

    record_id = zenodo_record["id"]

    resource_descriptors: list[Descriptor] = []
    for resource in dataset.resources or []:

        def _make_save_file(res: Resource) -> SaveFileCallback:
            def _save_file(props: SaveFileProps) -> str:
                stream = load_file_stream(props.normalized_path)
                file_data = stream.read()

                make_zenodo_api_request(
                    endpoint=f"/deposit/depositions/{record_id}/files",
                    method="POST",
                    upload=(props.denormalized_path, file_data),
                    api_key=api_key,
                    sandbox=sandbox,
                )

                return props.denormalized_path

            return _save_file

        resource_descriptors.append(
            save_resource_files(
                resource,
                basepath=basepath,
                with_remote=False,
                without_folders=True,
                save_file=_make_save_file(resource),
            )
        )

    denormalized = denormalize_dataset(dataset, basepath=basepath)
    descriptor: Descriptor = {
        **denormalized.model_dump(by_alias=True, exclude_none=True),
        "resources": resource_descriptors,
    }

    descriptor_bytes = stringify_descriptor(descriptor).encode()

    make_zenodo_api_request(
        endpoint=f"/deposit/depositions/{record_id}/files",
        method="POST",
        upload=("dataset.json", descriptor_bytes),
        api_key=api_key,
        sandbox=sandbox,
    )

    html_url = zenodo_record["links"]["html"]
    parsed = urllib.parse.urlparse(html_url)
    origin = f"{parsed.scheme}://{parsed.netloc}"

    return {
        "path": f"{origin}/records/{record_id}/files/dataset.json",
        "dataset_url": f"{origin}/uploads/{record_id}",
    }



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel

from .file import GithubFile
from .license import GithubLicense
from .owner import GithubOwner


class GithubRepository(FairspecModel):
    id: int | None = None
    name: str | None = None
    full_name: str | None = None
    owner: GithubOwner | None = None
    description: str | None = None
    created_at: str | None = None
    updated_at: str | None = None
    homepage: str | None = None
    size: int | None = None
    stargazers_count: int | None = None
    watchers_count: int | None = None
    language: str | None = None
    license: GithubLicense | None = None
    default_branch: str | None = None
    topics: list[str] | None = None
    private: bool | None = None
    archived: bool | None = None
    html_url: str | None = None
    git_url: str | None = None
    ssh_url: str | None = None
    clone_url: str | None = None
    files: list[GithubFile] | None = None



import pytest

from .basepath import get_common_local_basepath


class TestGetCommonLocalBasepath:
    @pytest.mark.parametrize(
        "description, paths, expected",
        [
            (
                "same directory different files",
                ["data/table1.csv", "data/table2.csv"],
                "data",
            ),
            (
                "nested directories",
                ["data/nested/file1.csv", "data/nested/file2.csv", "data/file3.csv"],
                "data",
            ),
            (
                "single path",
                ["data/file.csv"],
                "data",
            ),
            (
                "root level files",
                ["file1.csv", "file2.csv"],
                "",
            ),
            (
                "different top-level directories",
                ["data1/file1.csv", "data2/file2.csv"],
                "",
            ),
            (
                "empty paths array",
                [],
                None,
            ),
            (
                "some paths are remote",
                ["https://example.com/table.csv", "data/table.csv"],
                "data",
            ),
            (
                "all paths are remote",
                [
                    "https://example.com/table1.csv",
                    "https://example.com/table2.csv",
                ],
                None,
            ),
        ],
    )
    def test_get_common_local_basepath(self, description, paths, expected):
        result = get_common_local_basepath(paths)
        if expected is None:
            assert result is None
        else:
            assert result == expected



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class ZenodoCreator(FairspecModel):
    name: str | None = None
    affiliation: str | None = None
    identifiers: list[dict] | None = None



from fairspec_metadata import Resource

from fairspec_dataset.actions.file.temp import write_temp_file

from .infer import infer_bytes, infer_hash, infer_textual


class TestInferHash:
    def test_computes_sha256_hash_by_default(self):
        path = write_temp_file("Hello, World!")
        result = infer_hash(Resource(data=path))
        assert len(result) == 64

    def test_computes_md5_hash(self):
        path = write_temp_file("Hello, World!")
        result = infer_hash(Resource(data=path), hash_type="md5")
        assert len(result) == 32

    def test_computes_sha1_hash(self):
        path = write_temp_file("Hello, World!")
        result = infer_hash(Resource(data=path), hash_type="sha1")
        assert len(result) == 40

    def test_computes_sha512_hash(self):
        path = write_temp_file("Hello, World!")
        result = infer_hash(Resource(data=path), hash_type="sha512")
        assert len(result) == 128

    def test_consistent_hashes_for_same_content(self):
        path = write_temp_file("Hello, World!")
        result1 = infer_hash(Resource(data=path))
        result2 = infer_hash(Resource(data=path))
        assert result1 == result2


class TestInferBytes:
    def test_returns_file_size(self):
        path = write_temp_file("Hello, World!")
        result = infer_bytes(Resource(data=path))
        assert result == 13

    def test_handles_empty_files(self):
        path = write_temp_file("")
        result = infer_bytes(Resource(data=path))
        assert result == 0

    def test_handles_larger_files(self):
        path = write_temp_file("x" * 10000)
        result = infer_bytes(Resource(data=path))
        assert result == 10000

    def test_handles_binary_data(self):
        path = write_temp_file(bytes([0xFF, 0xD8, 0xFF, 0xE0]))
        result = infer_bytes(Resource(data=path))
        assert result == 4


class TestInferTextual:
    def test_returns_true_for_utf8_text(self):
        path = write_temp_file("Hello, World! This is UTF-8 text.")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_false_for_binary(self):
        path = write_temp_file(bytes([0xFF, 0xD8, 0xFF, 0xE0, 0x00]))
        assert infer_textual(Resource(data=path)) is False

    def test_uses_custom_sample_bytes(self):
        path = write_temp_file("This is a test file with UTF-8 content.")
        assert infer_textual(Resource(data=path), sample_bytes=20) is True

    def test_handles_large_text_files(self):
        path = write_temp_file("Hello World! " * 1000)
        assert infer_textual(Resource(data=path)) is True

    def test_handles_empty_files(self):
        path = write_temp_file("")
        assert infer_textual(Resource(data=path)) is True

    def test_handles_special_characters(self):
        path = write_temp_file("Special: é, ñ, ü, ö, à")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_true_for_ascii(self):
        path = write_temp_file(b"Simple ASCII text only")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_true_for_utf8_with_unicode(self):
        path = write_temp_file("Héllo, Wörld! 你好 مرحبا 🌍")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_true_for_cyrillic(self):
        path = write_temp_file("Привет мир")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_true_for_japanese(self):
        path = write_temp_file("こんにちは世界")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_true_for_arabic(self):
        path = write_temp_file("مرحبا بالعالم")
        assert infer_textual(Resource(data=path)) is True

    def test_returns_false_for_latin1(self):
        buffer = bytes(
            [
                0x43,
                0x61,
                0x66,
                0xE9,
                0x20,
                0x72,
                0xE9,
                0x73,
                0x75,
                0x6D,
                0xE9,
                0x20,
                0x6E,
                0x61,
                0xEF,
                0x76,
                0x65,
                0x20,
                0xE0,
                0x20,
                0x50,
                0x61,
                0x72,
                0x69,
                0x73,
                0x2E,
                0x20,
                0xC7,
                0x61,
                0x20,
                0x63,
                0x27,
                0x65,
                0x73,
                0x74,
                0x20,
                0x62,
                0x6F,
                0x6E,
                0x21,
            ]
        )
        path = write_temp_file(buffer)
        assert infer_textual(Resource(data=path)) is False

    def test_returns_false_for_windows_1252(self):
        buffer = bytes(
            [
                0x43,
                0x61,
                0x66,
                0xE9,
                0x20,
                0x6E,
                0x61,
                0xEF,
                0x76,
                0x65,
                0x20,
                0x72,
                0xE9,
                0x73,
                0x75,
                0x6D,
                0xE9,
            ]
        )
        path = write_temp_file(buffer)
        assert infer_textual(Resource(data=path)) is False



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from fairspec_metadata import get_basepath, get_data_paths, get_is_remote_path
from fairspec_metadata.actions.path.general import safe_relpath

if TYPE_CHECKING:
    from fairspec_metadata import Dataset


def get_dataset_basepath(dataset: Dataset) -> str | None:
    paths: list[str] = []

    for resource in dataset.resources or []:
        resource_paths = get_data_paths(resource)
        paths.extend(resource_paths)

    return get_common_local_basepath(paths)


def get_common_local_basepath(paths: list[str]) -> str | None:
    absolute_basepaths = [
        os.path.abspath(get_basepath(path))
        for path in paths
        if not get_is_remote_path(path)
    ]

    if not absolute_basepaths:
        return None

    segment_table = [
        [segment or "/" for segment in path.split(os.sep)]
        for path in absolute_basepaths
    ]

    column = 0
    segments: list[str] = []

    while True:
        segment_column = [
            row[column] if column < len(row) else None for row in segment_table
        ]
        unique_segments = set(segment_column)

        if len(unique_segments) != 1:
            break
        if segment_column[0] is None:
            break

        column += 1
        segments.append(segment_column[0])

    if not segments:
        raise ValueError("Cannot find common basepath")

    if segments[0].endswith(":"):
        segments[0] += os.sep

    basepath = safe_relpath(os.path.join(*segments))
    return "" if basepath == "." else basepath



from __future__ import annotations

import json
import os

from fairspec_metadata import (
    ContributorType,
    CreatorNameType,
    DateType,
    DescriptionType,
)
from fairspec_metadata import Contributor
from fairspec_metadata import Creator
from fairspec_metadata import DataciteDate
from fairspec_metadata import DataciteDescription
from fairspec_metadata import Rights
from fairspec_metadata import Subject
from fairspec_metadata import Title
from fairspec_metadata import Dataset
from fairspec_metadata import Integrity, IntegrityType
from fairspec_metadata import CsvFileDialect
from fairspec_metadata import Resource

from fairspec_dataset.plugins.ckan.models.dataset import CkanDataset
from .from_ckan import convert_dataset_from_ckan
from .to_ckan import convert_dataset_to_ckan


def _load_fixture() -> CkanDataset:
    path = os.path.join(os.path.dirname(__file__), "fixtures", "ckan-dataset.json")
    with open(path) as f:
        return CkanDataset(**json.load(f))


class TestConvertDatasetToCkan:
    def test_converts_fairspec_dataset_to_ckan_dataset(self):
        dataset = Dataset(
            titles=[Title(title="Test Package")],
            descriptions=[
                DataciteDescription(
                    description="This is a test package",
                    descriptionType=DescriptionType.Abstract,
                )
            ],
            version="1.0.0",
            rightsList=[
                Rights(
                    rights="Creative Commons Attribution",
                    rightsUri="http://www.opendefinition.org/licenses/cc-by",
                    rightsIdentifier="cc-by",
                )
            ],
            creators=[Creator(name="Test Author", nameType=CreatorNameType.Personal)],
            contributors=[
                Contributor(
                    name="Test Maintainer",
                    nameType=CreatorNameType.Personal,
                    contributorType=ContributorType.ContactPerson,
                )
            ],
            subjects=[
                Subject(subject="test"),
                Subject(subject="sample"),
                Subject(subject="data"),
            ],
            dates=[
                DataciteDate(date="2023-01-01T00:00:00Z", dateType=DateType.Created),
                DataciteDate(date="2023-01-02T00:00:00Z", dateType=DateType.Updated),
            ],
            resources=[
                Resource(
                    name="test_resource",
                    data="https://example.com/data.csv",
                    fileDialect=CsvFileDialect(),
                    descriptions=[
                        DataciteDescription(
                            description="Test resource",
                            descriptionType=DescriptionType.Abstract,
                        )
                    ],
                    integrity=Integrity(
                        type=IntegrityType.md5, hash="1234567890abcdef"
                    ),
                )
            ],
        )

        result = convert_dataset_to_ckan(dataset)

        assert result.title == "Test Package"
        assert result.notes == "This is a test package"
        assert result.version == "1.0.0"

        assert result.license_id == "cc-by"
        assert result.license_title == "Creative Commons Attribution"
        assert result.license_url == "http://www.opendefinition.org/licenses/cc-by"

        assert result.author == "Test Author"
        assert result.maintainer == "Test Maintainer"

        assert result.tags is not None
        assert len(result.tags) == 3
        for i, subject in enumerate(["test", "sample", "data"]):
            assert result.tags[i].name == subject
            assert result.tags[i].display_name == subject

        assert result.metadata_created == "2023-01-01T00:00:00Z"
        assert result.metadata_modified == "2023-01-02T00:00:00Z"

        assert result.resources is not None
        assert len(result.resources) == 1
        assert result.resources[0].name == "test_resource"
        assert result.resources[0].description == "Test resource"
        assert result.resources[0].hash == "1234567890abcdef"

    def test_handles_empty_resources_array(self):
        dataset = Dataset(resources=[])

        result = convert_dataset_to_ckan(dataset)

        assert result.resources == []

    def test_handles_undefined_optional_properties(self):
        dataset = Dataset(resources=[])

        result = convert_dataset_to_ckan(dataset)

        assert result.title is None
        assert result.notes is None
        assert result.version is None
        assert result.metadata_created is None
        assert result.metadata_modified is None
        assert result.license_id is None
        assert result.license_title is None
        assert result.license_url is None
        assert result.author is None
        assert result.maintainer is None
        assert result.tags == []
        assert result.resources == []

    def test_round_trip_ckan_to_dataset_to_ckan(self):
        original = _load_fixture()

        dataset = convert_dataset_from_ckan(original)
        result = convert_dataset_to_ckan(dataset)

        assert result.title == original.title
        assert result.notes == original.notes
        assert result.version == original.version

        assert result.license_id == original.license_id
        assert result.license_title == original.license_title
        assert result.license_url == original.license_url

        assert result.author == original.author
        assert result.maintainer == original.maintainer

        assert result.metadata_created == original.metadata_created
        assert result.metadata_modified == original.metadata_modified

        assert result.resources is not None
        assert len(result.resources) > 0

        assert result.tags is not None
        assert original.tags is not None
        assert len(result.tags) == len(original.tags)
        for original_tag in original.tags:
            matching = [t for t in result.tags if t.name == original_tag.name]
            assert len(matching) > 0



from __future__ import annotations

import json
import urllib.parse
import urllib.request
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


def make_ckan_api_request(
    *,
    ckan_url: str,
    action: str,
    payload: Descriptor,
    upload: tuple[str, bytes] | None = None,
    api_key: str | None = None,
) -> dict:
    parsed = urllib.parse.urlparse(ckan_url)
    url = f"{parsed.scheme}://{parsed.netloc}/api/3/action/{action}"

    headers: dict[str, str] = {}
    if api_key:
        headers["Authorization"] = api_key

    if upload:
        file_name, file_data = upload
        boundary = "----FormBoundary7MA4YWxkTrZu0gW"
        headers["Content-Type"] = f"multipart/form-data; boundary={boundary}"

        body = b""
        for key, value in payload.items():
            body += f"--{boundary}\r\n".encode()
            body += f'Content-Disposition: form-data; name="{key}"\r\n\r\n'.encode()
            body += f"{value}\r\n".encode()

        body += f"--{boundary}\r\n".encode()
        body += f'Content-Disposition: form-data; name="upload"; filename="{file_name}"\r\n'.encode()
        body += b"Content-Type: application/octet-stream\r\n\r\n"
        body += file_data
        body += f"\r\n--{boundary}--\r\n".encode()
    else:
        headers["Content-Type"] = "application/json"
        body = json.dumps(payload).encode()

    request = urllib.request.Request(url, data=body, headers=headers, method="POST")

    with urllib.request.urlopen(request) as response:
        data = json.loads(response.read().decode())

    if not data.get("success"):
        raise Exception(f"CKAN API error: {data.get('error')}")

    return data["result"]



from __future__ import annotations

from typing import TYPE_CHECKING, Unpack

from fairspec_metadata import load_dataset_descriptor
from fairspec_metadata import save_dataset_descriptor
from fairspec_metadata import get_file_extension, get_is_remote_path

from fairspec_dataset.models.dataset import SaveDatasetResult
from fairspec_dataset.plugin import DatasetPlugin

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor

    from fairspec_dataset.models.dataset import SaveDatasetOptions


class DescriptorPlugin(DatasetPlugin):
    def load_dataset(self, source: str) -> Descriptor | None:
        if not _get_is_json(source):
            return None
        dataset = load_dataset_descriptor(source)
        return dataset.model_dump(by_alias=True, exclude_none=True)

    def save_dataset(
        self, dataset: Dataset, **options: Unpack[SaveDatasetOptions]
    ) -> SaveDatasetResult | None:
        target = options["target"]
        if not _get_is_local_json(target):
            return None
        if not target.endswith("datapackage.json"):
            return None
        save_dataset_descriptor(dataset, path=target)
        return SaveDatasetResult(path=target)


def _get_is_json(path: str) -> bool:
    return get_file_extension(path) == "json"


def _get_is_local_json(path: str) -> bool:
    return _get_is_json(path) and not get_is_remote_path(path)



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import TableSchema

from fairspec_dataset.plugins.ckan.actions.table_schema.to_ckan import (
    convert_table_schema_to_ckan,
)
from fairspec_dataset.plugins.ckan.models.resource import CkanResource

if TYPE_CHECKING:
    from fairspec_metadata import Resource


def convert_resource_to_ckan(resource: Resource) -> CkanResource:
    name = resource.name if resource.name else None

    fmt = None
    file_dialect = resource.fileDialect
    if not isinstance(file_dialect, str) and file_dialect is not None:
        fmt = file_dialect.format
        if fmt:
            fmt = fmt.upper()

    description = None
    descriptions = resource.descriptions or []
    if descriptions and descriptions[0].description:
        description = descriptions[0].description

    hash_val = None
    integrity = resource.integrity
    if integrity and integrity.hash:
        hash_val = integrity.hash

    created = None
    updated = None
    dates = resource.dates or []
    created_date = next((d for d in dates if d.dateType == "Created"), None)
    if created_date:
        created = created_date.date
    updated_date = next((d for d in dates if d.dateType == "Updated"), None)
    if updated_date:
        updated = updated_date.date

    schema = None
    table_schema = resource.tableSchema
    if isinstance(table_schema, TableSchema):
        schema = convert_table_schema_to_ckan(table_schema)

    return CkanResource(
        name=name,
        format=fmt,
        description=description,
        hash=hash_val,
        created=created,
        last_modified=updated,
        schema_=schema,
    )



from __future__ import annotations

import urllib.request
from io import BytesIO
from typing import BinaryIO

from fairspec_metadata import ResourceDataPath, get_is_remote_path


def load_file_stream(
    data_path: ResourceDataPath,
    *,
    index: int = 0,
    max_bytes: int | None = None,
) -> BinaryIO:
    paths = data_path if isinstance(data_path, list) else [data_path]

    if index >= len(paths) or index < 0:
        raise ValueError(
            f"Cannot stream resource {paths[index] if index < len(paths) else None} at index {index}"
        )

    path = paths[index]
    is_remote = get_is_remote_path(path)

    if is_remote:
        return _load_remote_file_stream(path, max_bytes=max_bytes)

    return _load_local_file_stream(path, max_bytes=max_bytes)


def _load_local_file_stream(
    path: str,
    *,
    max_bytes: int | None = None,
) -> BinaryIO:
    if max_bytes is not None:
        with open(path, "rb") as f:
            data = f.read(max_bytes)
        return BytesIO(data)

    return open(path, "rb")


def _load_remote_file_stream(
    path: str,
    *,
    max_bytes: int | None = None,
) -> BinaryIO:
    with urllib.request.urlopen(path) as response:
        if max_bytes is not None:
            data = response.read(max_bytes)
        else:
            data = response.read()

    return BytesIO(data)



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel

from .file import ZenodoFile
from .link import ZenodoLink
from .metadata import ZenodoMetadata


class ZenodoRecord(FairspecModel):
    id: int | None = None
    links: ZenodoLink | None = None
    metadata: ZenodoMetadata | None = None
    files: list[ZenodoFile] | None = None
    state: str | None = None
    submitted: bool | None = None



import os

import pytest
from fairspec_metadata import DescriptionType
from fairspec_metadata import DataciteDescription
from fairspec_metadata import Subject
from fairspec_metadata import Title
from fairspec_metadata import Dataset
from fairspec_metadata import CsvFileDialect
from fairspec_metadata import Resource
from fairspec_metadata import IntegerColumnProperty
from fairspec_metadata import StringColumnProperty
from fairspec_metadata import TableSchema

from fairspec_dataset.actions.file.temp import get_temp_file_path, write_temp_file
from .load import load_dataset_from_zip
from .save import save_dataset_to_zip


class TestSaveDatasetToZip:
    def test_saves_basic_dataset_to_zip(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(resources=[Resource(name="test_res", data=[{"id": 1}])])

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)
        assert os.path.getsize(path) > 0

    def test_saves_dataset_with_metadata(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            titles=[Title(title="Test Dataset")],
            descriptions=[
                DataciteDescription(
                    description="A test", descriptionType=DescriptionType.Abstract
                )
            ],
            version="1.0",
            resources=[Resource(name="test_res", data=[{"id": 1}])],
        )

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)
        assert os.path.getsize(path) > 0

    def test_saves_dataset_with_inline_data_resources(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[Resource(name="test_res", data=[{"id": 1}, {"id": 2}])]
        )

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)

    def test_saves_dataset_with_file_resources(self):
        csv_path = write_temp_file("id,name\n1,alice\n2,bob\n", format="csv")
        path = get_temp_file_path(format="zip")
        dataset = Dataset(resources=[Resource(name="test_res", data=csv_path)])

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)

    def test_saves_dataset_with_multiple_resources(self):
        csv_path = write_temp_file("id,name\n1,alice\n", format="csv")
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[
                Resource(name="file_res", data=csv_path),
                Resource(name="inline_res", data=[{"id": 1}]),
            ]
        )

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)

    def test_saves_dataset_with_table_schema(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[
                Resource(
                    name="test_res",
                    data=[{"id": 1}],
                    tableSchema=TableSchema(
                        properties={
                            "id": IntegerColumnProperty(),
                            "name": StringColumnProperty(),
                        }
                    ),
                )
            ]
        )

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)

    def test_saves_dataset_with_dialect(self):
        csv_path = write_temp_file("id;name\n1;alice\n", format="csv")
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[
                Resource(
                    name="test_res",
                    data=csv_path,
                    fileDialect=CsvFileDialect(delimiter=";"),
                )
            ]
        )

        save_dataset_to_zip(dataset, archive_path=path)

        assert os.path.exists(path)

    def test_roundtrip_preserves_structure(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            titles=[Title(title="My Dataset")],
            descriptions=[
                DataciteDescription(
                    description="Desc", descriptionType=DescriptionType.Abstract
                )
            ],
            resources=[Resource(name="test_res", data=[{"id": 1}])],
        )
        save_dataset_to_zip(dataset, archive_path=path)

        result = load_dataset_from_zip(path)

        assert result.titles is not None
        assert result.titles[0].title == "My Dataset"
        assert result.descriptions is not None
        assert result.descriptions[0].description == "Desc"
        assert result.resources is not None
        assert len(result.resources) == 1

    def test_roundtrip_preserves_metadata(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            titles=[Title(title="My Dataset")],
            version="2.0",
            subjects=[Subject(subject="science")],
            resources=[Resource(name="test_res", data=[{"id": 1}])],
        )
        save_dataset_to_zip(dataset, archive_path=path)

        result = load_dataset_from_zip(path)

        assert result.titles is not None
        assert result.titles[0].title == "My Dataset"
        assert result.version == "2.0"
        assert result.subjects is not None
        assert result.subjects[0].subject == "science"

    def test_roundtrip_preserves_table_schema(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[
                Resource(
                    name="test_res",
                    data=[{"id": 1}],
                    tableSchema=TableSchema(
                        properties={
                            "id": IntegerColumnProperty(),
                            "name": StringColumnProperty(),
                        }
                    ),
                )
            ]
        )
        save_dataset_to_zip(dataset, archive_path=path)

        result = load_dataset_from_zip(path)

        assert result.resources is not None
        assert isinstance(result.resources[0].tableSchema, TableSchema)
        assert result.resources[0].tableSchema.properties is not None
        assert len(result.resources[0].tableSchema.properties) == 2

    def test_roundtrip_preserves_file_resources(self):
        csv_path = write_temp_file("id,name\n1,alice\n2,bob\n", format="csv")
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[
                Resource(
                    name="test_res",
                    data=csv_path,
                    fileDialect=CsvFileDialect(),
                )
            ]
        )
        save_dataset_to_zip(dataset, archive_path=path)

        result = load_dataset_from_zip(path)

        assert result.resources is not None
        assert result.resources[0].fileDialect is not None

    def test_throws_error_for_existing_file(self):
        path = write_temp_file("existing content", format="zip")
        dataset = Dataset(resources=[Resource(name="test_res", data=[{"id": 1}])])

        with pytest.raises(FileExistsError):
            save_dataset_to_zip(dataset, archive_path=path)

    def test_creates_valid_zip_structure(self):
        path = get_temp_file_path(format="zip")
        dataset = Dataset(resources=[Resource(name="test_res", data=[{"id": 1}])])
        save_dataset_to_zip(dataset, archive_path=path)

        result = load_dataset_from_zip(path)

        assert result.resources is not None
        assert result.resources[0].name == "test_res"

    def test_roundtrip_with_multiple_file_resources(self):
        csv1 = write_temp_file("id,name\n1,alice\n", filename="data1.csv")
        csv2 = write_temp_file("id,name\n2,bob\n", filename="data2.csv")
        path = get_temp_file_path(format="zip")
        dataset = Dataset(
            resources=[
                Resource(name="first_res", data=csv1),
                Resource(name="second_res", data=csv2),
            ]
        )
        save_dataset_to_zip(dataset, archive_path=path)

        result = load_dataset_from_zip(path)

        assert result.resources is not None
        assert len(result.resources) == 2
        assert result.resources[0].name == "first_res"
        assert result.resources[1].name == "second_res"



from unittest.mock import MagicMock, patch

from .plugin import CkanPlugin


class TestLoadDataset:
    def setup_method(self):
        self.plugin = CkanPlugin()

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_loads_from_ckan_url(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset("https://demo.ckan.org/dataset/my-dataset")
        mock_load.assert_called_once_with("https://demo.ckan.org/dataset/my-dataset")
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_returns_none_for_url_without_dataset_path(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://demo.ckan.org/about")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_returns_none_for_local_paths(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("/tmp/data")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_returns_none_for_github_urls(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://github.com/owner/repo")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_handles_additional_path_segments(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset(
            "https://demo.ckan.org/en_GB/dataset/my-dataset"
        )
        mock_load.assert_called_once()
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_handles_query_parameters(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset(
            "https://demo.ckan.org/dataset/my-dataset?page=1"
        )
        mock_load.assert_called_once()
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_handles_http_urls(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset("http://demo.ckan.org/dataset/my-dataset")
        mock_load.assert_called_once()
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_returns_none_for_zenodo_urls(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://zenodo.org/records/123456")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.ckan.plugin.load_dataset_from_ckan")
    def test_returns_none_for_dataset_in_query_params_only(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://example.com/search?name=dataset")
        mock_load.assert_not_called()
        assert result is None



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fairspec_metadata import Dataset


def convert_dataset_to_zenodo(dataset: Dataset) -> dict:
    metadata: dict = {"upload_type": "dataset"}

    titles = dataset.titles or []
    if titles and titles[0].title:
        metadata["title"] = titles[0].title

    descriptions = dataset.descriptions or []
    if descriptions and descriptions[0].description:
        metadata["description"] = descriptions[0].description
    elif titles and titles[0].title:
        metadata["description"] = titles[0].title
    else:
        metadata["description"] = "Dataset created with fairspec"

    if dataset.version:
        metadata["version"] = dataset.version

    rights_list = dataset.rightsList or []
    if rights_list:
        rights = rights_list[0]
        if rights.rights:
            metadata["license"] = rights.rights

    creators = dataset.creators or []
    if creators:
        metadata["creators"] = []
        for creator in creators:
            zenodo_creator: dict = {"name": creator.name}
            affiliations = creator.affiliation or []
            if affiliations and affiliations[0].name:
                zenodo_creator["affiliation"] = affiliations[0].name
            metadata["creators"].append(zenodo_creator)
    else:
        metadata["creators"] = [
            {"name": "Unknown Author", "affiliation": "Unknown Affiliation"}
        ]

    subjects = dataset.subjects or []
    if subjects:
        metadata["keywords"] = [s.subject for s in subjects]

    dates = dataset.dates or []
    if dates:
        issued = next((d for d in dates if d.dateType == "Issued"), None)
        if issued:
            metadata["publication_date"] = issued.date

    if dataset.doi:
        metadata["doi"] = dataset.doi

    return {"metadata": metadata}



import os
from io import BytesIO

import pytest

from .save import save_file_stream


class TestSaveFileStream:
    def test_saves_stream_to_file(self, tmp_path):
        path = str(tmp_path / "test.txt")
        save_file_stream(BytesIO(b"Hello, World!"), path=path)
        assert os.path.exists(path)
        with open(path, "rb") as f:
            assert f.read() == b"Hello, World!"

    def test_saves_stream_with_multiple_chunks(self, tmp_path):
        path = str(tmp_path / "chunks.txt")
        save_file_stream(BytesIO(b"Hello, World!"), path=path)
        with open(path, "rb") as f:
            assert f.read() == b"Hello, World!"

    def test_creates_nested_directories(self, tmp_path):
        path = str(tmp_path / "nested" / "dir" / "file.txt")
        save_file_stream(BytesIO(b"Nested content"), path=path)
        with open(path, "rb") as f:
            assert f.read() == b"Nested content"

    def test_raises_when_file_exists_and_overwrite_false(self, tmp_path):
        path = str(tmp_path / "existing.txt")
        save_file_stream(BytesIO(b"Initial content"), path=path)
        with pytest.raises(FileExistsError):
            save_file_stream(BytesIO(b"New content"), path=path, overwrite=False)

    def test_raises_when_file_exists_and_overwrite_not_specified(self, tmp_path):
        path = str(tmp_path / "existing2.txt")
        save_file_stream(BytesIO(b"Initial content"), path=path)
        with pytest.raises(FileExistsError):
            save_file_stream(BytesIO(b"New content"), path=path)

    def test_overwrites_when_overwrite_true(self, tmp_path):
        path = str(tmp_path / "overwrite.txt")
        save_file_stream(BytesIO(b"Initial content"), path=path)
        save_file_stream(BytesIO(b"New content"), path=path, overwrite=True)
        with open(path, "rb") as f:
            assert f.read() == b"New content"

    def test_saves_binary_data(self, tmp_path):
        binary_data = bytes([0x00, 0x01, 0x02, 0x03, 0xFF])
        path = str(tmp_path / "binary.bin")
        save_file_stream(BytesIO(binary_data), path=path)
        with open(path, "rb") as f:
            assert f.read() == binary_data

    def test_saves_empty_stream(self, tmp_path):
        path = str(tmp_path / "empty.txt")
        save_file_stream(BytesIO(b""), path=path)
        with open(path, "rb") as f:
            assert f.read() == b""

    def test_saves_large_stream(self, tmp_path):
        content = b"A" * 10000
        path = str(tmp_path / "large.txt")
        save_file_stream(BytesIO(content), path=path)
        with open(path, "rb") as f:
            assert f.read() == content

    def test_saves_unicode_characters(self, tmp_path):
        content = "Unicode: 你好世界 🌍 مرحبا".encode()
        path = str(tmp_path / "unicode.txt")
        save_file_stream(BytesIO(content), path=path)
        with open(path, "rb") as f:
            assert f.read() == content



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class ZenodoLink(FairspecModel):
    self: str | None = None
    html: str | None = None
    files: str | None = None
    bucket: str | None = None
    publish: str | None = None
    discard: str | None = None
    edit: str | None = None



from __future__ import annotations

import json
import os

from fairspec_dataset.plugins.ckan.models.field import CkanField
from fairspec_dataset.plugins.ckan.models.schema import CkanSchema
from .from_ckan import convert_table_schema_from_ckan


def _load_fixture() -> CkanSchema:
    path = os.path.join(os.path.dirname(__file__), "fixtures", "ckan-schema.json")
    with open(path) as f:
        return CkanSchema(**json.load(f))


class TestConvertTableSchemaFromCkan:
    def test_converts_ckan_schema_to_fairspec_table_schema(self):
        ckan_schema = _load_fixture()

        result = convert_table_schema_from_ckan(ckan_schema)

        assert result.properties is not None
        properties = result.properties
        assert ckan_schema.fields is not None
        assert len(properties) == len(ckan_schema.fields)

        id_col = properties["id"]
        assert id_col.type == "integer"
        assert id_col.title == "ID"
        assert id_col.description == "Unique identifier"

        name_col = properties["name"]
        assert name_col.type == "string"
        assert name_col.title == "Name"
        assert name_col.description == "Person's full name"

        age_col = properties["age"]
        assert age_col.type == "integer"
        assert age_col.title is None
        assert age_col.description is None

        score_col = properties["score"]
        assert score_col.type == "number"
        assert score_col.title == "Score"
        assert score_col.description == "Test score"

        is_active_col = properties["is_active"]
        assert is_active_col.type == "boolean"

        birth_date_col = properties["birth_date"]
        assert birth_date_col.type == "string"
        assert birth_date_col.format == "date"
        assert birth_date_col.title == "Birth Date"
        assert birth_date_col.description == "Date of birth"

        start_time_col = properties["start_time"]
        assert start_time_col.type == "string"
        assert start_time_col.format == "time"

        created_at_col = properties["created_at"]
        assert created_at_col.type == "string"
        assert created_at_col.format == "date-time"
        assert created_at_col.title == "Created At"
        assert created_at_col.description == "Timestamp when record was created"

        metadata_col = properties["metadata"]
        assert metadata_col.type == "object"

        tags_col = properties["tags"]
        assert tags_col.type == "array"
        assert tags_col.title == "Tags"
        assert tags_col.description == "List of tags"

    def test_converts_ckan_type_aliases(self):
        ckan_schema = _load_fixture()

        result = convert_table_schema_from_ckan(ckan_schema)
        assert result.properties is not None
        properties = result.properties

        assert properties["string_field"].type == "string"
        assert properties["integer_field"].type == "integer"
        assert properties["number_field"].type == "number"
        assert properties["float_field"].type == "number"
        assert properties["boolean_field"].type == "boolean"
        assert properties["datetime_field"].type == "string"
        assert properties["datetime_field"].format == "date-time"
        assert properties["object_field"].type == "object"

    def test_handles_unknown_field_types(self):
        ckan_schema = _load_fixture()

        result = convert_table_schema_from_ckan(ckan_schema)
        assert result.properties is not None
        properties = result.properties

        assert properties["unknown_field"].type == "string"

    def test_respects_type_override_in_field_info(self):
        ckan_schema = _load_fixture()

        result = convert_table_schema_from_ckan(ckan_schema)
        assert result.properties is not None
        properties = result.properties

        override_col = properties["override_field"]
        assert override_col.type == "integer"
        assert override_col.title == "Override Field"
        assert override_col.description == "Field with type override"

    def test_handles_empty_fields_array(self):
        ckan_schema = CkanSchema(fields=[])

        result = convert_table_schema_from_ckan(ckan_schema)

        assert result.properties is None or len(result.properties) == 0

    def test_handles_fields_without_info_object(self):
        ckan_schema = CkanSchema(fields=[CkanField(id="simple_field", type="text")])

        result = convert_table_schema_from_ckan(ckan_schema)
        assert result.properties is not None
        properties = result.properties

        assert len(properties) == 1
        col = properties["simple_field"]
        assert col.type == "string"
        assert col.title is None
        assert col.description is None

    def test_handles_case_insensitive_type_conversion(self):
        ckan_schema = CkanSchema(
            fields=[
                CkanField(id="field1", type="TEXT"),
                CkanField(id="field2", type="INT"),
                CkanField(id="field3", type="BOOL"),
                CkanField(id="field4", type="TIMESTAMP"),
            ],
        )

        result = convert_table_schema_from_ckan(ckan_schema)
        assert result.properties is not None
        properties = result.properties

        assert properties["field1"].type == "string"
        assert properties["field2"].type == "integer"
        assert properties["field3"].type == "boolean"
        assert properties["field4"].type == "string"
        assert properties["field4"].format == "date-time"



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import (
    IntegrityError,
    Report,
    TextualError,
    create_report,
)
from fairspec_metadata import FairspecError

from .infer import infer_hash, infer_textual

if TYPE_CHECKING:
    from fairspec_metadata import Resource


def validate_file(resource: Resource) -> Report:
    errors: list[FairspecError] = []

    if resource.textual:
        actual_textual = infer_textual(resource)

        if not actual_textual:
            errors.append(TextualError(type="file/textual"))

    integrity = resource.integrity
    if integrity:
        expected_hash = integrity.hash
        actual_hash = infer_hash(resource, hash_type=integrity.type)

        if actual_hash != expected_hash:
            errors.append(
                IntegrityError(
                    type="file/integrity",
                    hashType=integrity.type,
                    expectedHash=expected_hash,
                    actualHash=actual_hash or "",
                )
            )

    return create_report(errors)



from __future__ import annotations

import atexit
import os
import tempfile


def write_temp_file(
    content: str | bytes,
    *,
    persist: bool = False,
    filename: str | None = None,
    format: str | None = None,
) -> str:
    path = get_temp_file_path(persist=persist, filename=filename, format=format)
    mode = "wb" if isinstance(content, bytes) else "w"
    encoding = None if isinstance(content, bytes) else "utf-8"
    newline = None if isinstance(content, bytes) else ""
    with open(path, mode, encoding=encoding, newline=newline) as f:
        f.write(content)
    return path


def get_temp_file_path(
    *,
    persist: bool = False,
    filename: str | None = None,
    format: str | None = None,
) -> str:
    if filename:
        dir_path = tempfile.mkdtemp()
        path = os.path.join(dir_path, filename)
    else:
        suffix = f".{format}" if format else ""
        fd, path = tempfile.mkstemp(suffix=suffix)
        os.close(fd)
        os.unlink(path)

    if not persist:
        atexit.register(_cleanup_file, path)

    return path


def _cleanup_file(path: str) -> None:
    try:
        os.unlink(path)
    except OSError:
        pass



from .actions.dataset.load import load_dataset_from_zenodo
from .actions.dataset.save import save_dataset_to_zenodo
from .plugin import ZenodoPlugin

__all__ = [
    "ZenodoPlugin",
    "load_dataset_from_zenodo",
    "save_dataset_to_zenodo",
]



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class CkanOrganization(FairspecModel):
    id: str | None = None
    name: str | None = None
    title: str | None = None
    description: str | None = None



import pytest

from fairspec_dataset.actions.file.temp import write_temp_file

from .load import load_file_stream


class TestLoadFileStream:
    def test_loads_stream_from_single_local_file(self):
        path = write_temp_file("Hello, World!")
        stream = load_file_stream(path)
        assert stream.read() == b"Hello, World!"

    def test_loads_stream_from_array_using_default_index(self):
        file1 = write_temp_file("First file content")
        file2 = write_temp_file("Second file content")
        stream = load_file_stream([file1, file2])
        assert stream.read() == b"First file content"

    def test_loads_stream_from_array_using_specified_index(self):
        file1 = write_temp_file("First file content")
        file2 = write_temp_file("Second file content")
        stream = load_file_stream([file1, file2], index=1)
        assert stream.read() == b"Second file content"

    def test_limits_stream_to_max_bytes(self):
        path = write_temp_file("This is a long content that should be truncated")
        stream = load_file_stream(path, max_bytes=10)
        data = stream.read()
        assert data == b"This is a "
        assert len(data) == 10

    def test_raises_error_for_invalid_index(self):
        path = write_temp_file("content")
        with pytest.raises(ValueError, match="Cannot stream resource"):
            load_file_stream([path], index=5)

    def test_raises_error_for_empty_array(self):
        with pytest.raises(ValueError, match="Cannot stream resource"):
            load_file_stream([], index=0)

    def test_handles_large_files(self):
        content = "A" * 10000
        path = write_temp_file(content)
        stream = load_file_stream(path)
        data = stream.read()
        assert len(data) == 10000
        assert data == content.encode()

    def test_handles_binary_content(self):
        binary_data = bytes([0x00, 0x01, 0x02, 0x03, 0xFF])
        path = write_temp_file(binary_data)
        stream = load_file_stream(path)
        assert stream.read() == binary_data

    def test_handles_empty_files(self):
        path = write_temp_file("")
        stream = load_file_stream(path)
        assert stream.read() == b""

    def test_limits_bytes_correctly(self):
        path = write_temp_file("0123456789ABCDEFGHIJ")
        stream = load_file_stream(path, max_bytes=5)
        assert stream.read() == b"01234"

    def test_handles_max_bytes_larger_than_file(self):
        path = write_temp_file("Short")
        stream = load_file_stream(path, max_bytes=1000)
        assert stream.read() == b"Short"



from __future__ import annotations

import urllib.parse
from collections.abc import Callable
from typing import TYPE_CHECKING

from fairspec_metadata import (
    denormalize_dataset,
    get_file_extension,
    stringify_descriptor,
)

from fairspec_dataset.actions.dataset.basepath import get_dataset_basepath
from fairspec_dataset.actions.resource.save import SaveFileProps, save_resource_files
from fairspec_dataset.actions.stream.load import load_file_stream

from fairspec_dataset.plugins.ckan.actions.resource.to_ckan import (
    convert_resource_to_ckan,
)
from fairspec_dataset.plugins.ckan.services.ckan import make_ckan_api_request
from .to_ckan import convert_dataset_to_ckan

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor
    from fairspec_metadata import Resource


def save_dataset_to_ckan(
    dataset: Dataset,
    *,
    api_key: str,
    ckan_url: str,
    owner_org: str,
    dataset_name: str,
) -> dict:
    basepath = get_dataset_basepath(dataset)
    ckan_dataset = convert_dataset_to_ckan(dataset)

    payload = {
        **ckan_dataset.model_dump(by_alias=True, exclude_none=True),
        "name": dataset_name,
        "owner_org": owner_org,
        "resources": [],
    }

    result = make_ckan_api_request(
        action="package_create",
        payload=payload,
        ckan_url=ckan_url,
        api_key=api_key,
    )

    parsed = urllib.parse.urlparse(ckan_url)
    dataset_url = f"{parsed.scheme}://{parsed.netloc}/dataset/{result['name']}"

    resource_descriptors: list[Descriptor] = []
    for resource in dataset.resources or []:

        def _make_save_file(res: Resource) -> Callable[[SaveFileProps], str]:
            def _save_file(props: SaveFileProps) -> str:
                ckan_resource = convert_resource_to_ckan(
                    res.model_dump(by_alias=True, exclude_none=True)
                )
                extension = get_file_extension(props.normalized_path)

                upload_payload: dict = {
                    **ckan_resource.model_dump(by_alias=True, exclude_none=True),
                    "package_id": dataset_name,
                    "name": props.denormalized_path,
                }
                if extension:
                    upload_payload["format"] = extension.upper()

                stream = load_file_stream(props.normalized_path)
                file_data = stream.read()

                upload_result = make_ckan_api_request(
                    action="resource_create",
                    payload=upload_payload,
                    upload=(props.denormalized_path, file_data),
                    ckan_url=ckan_url,
                    api_key=api_key,
                )

                return upload_result["url"]

            return _save_file

        resource_descriptors.append(
            save_resource_files(
                resource,
                basepath=basepath,
                with_remote=True,
                without_folders=True,
                save_file=_make_save_file(resource),
            )
        )

    denormalized = denormalize_dataset(dataset, basepath=basepath)
    descriptor: Descriptor = {
        **denormalized.model_dump(by_alias=True, exclude_none=True),
        "resources": resource_descriptors,
    }

    descriptor_bytes = stringify_descriptor(descriptor).encode()

    make_ckan_api_request(
        action="resource_create",
        payload={
            "package_id": dataset_name,
            "name": "datapackage.json",
        },
        upload=("datapackage.json", descriptor_bytes),
        ckan_url=ckan_url,
        api_key=api_key,
    )

    return {
        "path": result.get("url"),
        "dataset_url": dataset_url,
    }



from __future__ import annotations

from fairspec_dataset.actions.stream.load import load_file_stream


def load_file(path: str, *, max_bytes: int | None = None) -> bytes:
    stream = load_file_stream(path, max_bytes=max_bytes)
    return stream.read()



import os

from .create import create_folder


class TestCreateFolder:
    def test_creates_simple_folder(self, tmp_path):
        path = str(tmp_path / "test-folder")
        create_folder(path)
        assert os.path.isdir(path)

    def test_creates_nested_folders(self, tmp_path):
        path = str(tmp_path / "parent" / "child" / "grandchild")
        create_folder(path)
        assert os.path.isdir(path)
        assert os.path.isdir(str(tmp_path / "parent"))
        assert os.path.isdir(str(tmp_path / "parent" / "child"))

    def test_no_error_when_folder_exists(self, tmp_path):
        path = str(tmp_path / "existing")
        create_folder(path)
        create_folder(path)
        assert os.path.isdir(path)

    def test_creates_deeply_nested_directories(self, tmp_path):
        path = str(tmp_path / "l1" / "l2" / "l3" / "l4" / "l5")
        create_folder(path)
        assert os.path.isdir(path)

    def test_creates_folder_with_special_characters(self, tmp_path):
        path = str(tmp_path / "folder-with_special.chars")
        create_folder(path)
        assert os.path.isdir(path)

    def test_creates_sibling_folders(self, tmp_path):
        for name in ("folder1", "folder2", "folder3"):
            create_folder(str(tmp_path / name))
        for name in ("folder1", "folder2", "folder3"):
            assert os.path.isdir(str(tmp_path / name))

    def test_creates_complex_directory_structure(self, tmp_path):
        dirs = [
            str(tmp_path / "project" / "src" / "components"),
            str(tmp_path / "project" / "src" / "utils"),
            str(tmp_path / "project" / "tests" / "unit"),
            str(tmp_path / "project" / "tests" / "integration"),
        ]
        for d in dirs:
            create_folder(d)
        for d in dirs:
            assert os.path.isdir(d)



import pytest
from fairspec_metadata import Resource

from .save import SaveFileProps, save_resource_files


def _identity_save(props: SaveFileProps) -> str:
    return props.denormalized_path


class TestSaveResourceFiles:
    @pytest.mark.parametrize(
        "description, basepath, resource, expected, with_remote, without_folders",
        [
            (
                "local path",
                "data",
                {
                    "data": "data/table.csv",
                    "dataSchema": "data/data-schema.json",
                    "tableSchema": "data/table-schema.json",
                },
                {
                    "data": "table.csv",
                    "dataSchema": "data-schema.json",
                    "tableSchema": "table-schema.json",
                },
                False,
                False,
            ),
            (
                "local paths",
                "data",
                {"data": ["data/table1.csv", "data/table2.csv"]},
                {"data": ["table1.csv", "table2.csv"]},
                False,
                False,
            ),
            (
                "local path and remote path",
                "data",
                {
                    "data": "data/table.csv",
                    "tableSchema": "https://example.com/schema.json",
                },
                {
                    "data": "table.csv",
                    "tableSchema": "https://example.com/schema.json",
                },
                False,
                False,
            ),
            (
                "local path and remote path using with_remote",
                "data",
                {
                    "data": "data/table.csv",
                    "tableSchema": "https://example.com/schema.json",
                },
                {
                    "data": "table.csv",
                    "tableSchema": "schema.json",
                },
                True,
                False,
            ),
            (
                "remote paths with same filename using with_remote",
                "data",
                {
                    "data": [
                        "http://example1.com/table.csv",
                        "http://example2.com/table.csv",
                        "http://example3.com/table.csv",
                        "http://example4.com/table.csv.zip",
                        "http://example5.com/table.csv.zip",
                    ],
                },
                {
                    "data": [
                        "table.csv",
                        "table-1.csv",
                        "table-2.csv",
                        "table.csv.zip",
                        "table-1.csv.zip",
                    ],
                },
                True,
                False,
            ),
            (
                "local paths in different folders",
                "data",
                {
                    "data": "data/folder1/table.csv",
                    "tableSchema": "data/folder2/schema.json",
                },
                {
                    "data": "folder1/table.csv",
                    "tableSchema": "folder2/schema.json",
                },
                False,
                False,
            ),
            (
                "local paths in different folders using without_folders",
                "data",
                {
                    "data": "data/folder1/table.csv",
                    "tableSchema": "data/folder2/schema.json",
                },
                {
                    "data": "folder1-table.csv",
                    "tableSchema": "folder2-schema.json",
                },
                False,
                True,
            ),
        ],
    )
    def test_save_resource_files(
        self,
        description,
        basepath,
        resource,
        expected,
        with_remote,
        without_folders,
    ):
        result = save_resource_files(
            Resource(**resource),
            basepath=basepath,
            with_remote=with_remote,
            without_folders=without_folders,
            save_file=_identity_save,
        )
        assert result == expected



from __future__ import annotations

import os
import zipfile

from fairspec_metadata import load_dataset_descriptor
from fairspec_metadata import Dataset

from fairspec_dataset.actions.folder.temp import get_temp_folder_path


def load_dataset_from_zip(archive_path: str) -> Dataset:
    basepath = get_temp_folder_path()

    with zipfile.ZipFile(archive_path, "r") as zf:
        zf.extractall(basepath)

    return load_dataset_descriptor(os.path.join(basepath, "dataset.json"))



from __future__ import annotations

from fairspec_metadata.models.base import FairspecModel


class ZenodoFileLinks(FairspecModel):
    self: str | None = None


class ZenodoFile(FairspecModel):
    id: str | None = None
    key: str | None = None
    size: int | None = None
    checksum: str | None = None
    links: ZenodoFileLinks | None = None



from __future__ import annotations

from fairspec_metadata import load_dataset_descriptor
from fairspec_metadata import Dataset


def merge_datasets(
    *,
    system_dataset: Dataset,
    user_dataset_path: str | None = None,
) -> Dataset:
    system = system_dataset.model_dump(by_alias=True, exclude_none=True)

    user_dataset = (
        load_dataset_descriptor(user_dataset_path).model_dump(
            by_alias=True, exclude_none=True
        )
        if user_dataset_path
        else None
    )

    merged = {**system, **user_dataset} if user_dataset else {**system}
    return Dataset(**merged)



from __future__ import annotations

import os

import pytest


@pytest.fixture(scope="module")
def vcr_cassette_dir(request):
    return os.path.join(os.path.dirname(request.fspath), "fixtures", "generated")



from __future__ import annotations

from io import BytesIO

from fairspec_dataset.actions.stream.save import save_file_stream


def save_file(path: str, data: bytes, *, overwrite: bool = False) -> None:
    save_file_stream(BytesIO(data), path=path, overwrite=overwrite)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from fairspec_metadata import get_file_name
from fairspec_metadata import DateType, DescriptionType
from fairspec_metadata import DataciteDate
from fairspec_metadata import DataciteDescription
from fairspec_metadata import Integrity, IntegrityType
from fairspec_metadata import Resource

from fairspec_dataset.plugins.ckan.actions.table_schema.from_ckan import (
    convert_table_schema_from_ckan,
)

if TYPE_CHECKING:
    from fairspec_dataset.plugins.ckan.models.resource import CkanResource


def convert_resource_from_ckan(ckan_resource: CkanResource) -> Resource:
    name = _convert_name(ckan_resource.name) if ckan_resource.name else None

    descriptions = (
        [
            DataciteDescription(
                description=ckan_resource.description,
                descriptionType=DescriptionType.Abstract,
            )
        ]
        if ckan_resource.description
        else None
    )

    sizes = [f"{ckan_resource.size} bytes"] if ckan_resource.size else None

    integrity = (
        Integrity(type=IntegrityType.md5, hash=ckan_resource.hash)
        if ckan_resource.hash
        else None
    )

    dates: list[DataciteDate] = []
    if ckan_resource.created:
        dates.append(
            DataciteDate(date=ckan_resource.created, dateType=DateType.Created)
        )
    if ckan_resource.last_modified:
        dates.append(
            DataciteDate(date=ckan_resource.last_modified, dateType=DateType.Updated)
        )

    table_schema = (
        convert_table_schema_from_ckan(ckan_resource.schema_)
        if ckan_resource.schema_
        else None
    )

    return Resource(
        data=ckan_resource.url or "",
        name=name,
        descriptions=descriptions,
        sizes=sizes,
        integrity=integrity,
        dates=dates if dates else None,
        tableSchema=table_schema,
        unstable_customMetadata={
            "ckanKey": get_file_name(ckan_resource.url or ""),
            "ckanUrl": ckan_resource.url or "",
            "ckanId": ckan_resource.id,
        },
    )


def _convert_name(name: str) -> str:
    result = re.sub(r"[\s.()/\\,\-]+", "_", name).lower()
    result = re.sub(r"[^a-z0-9_]", "", result)
    result = re.sub(r"^(\d)", r"_\1", result)
    return result[:100]



from __future__ import annotations

import hashlib
import os
from typing import TYPE_CHECKING

from charset_normalizer import from_bytes

from fairspec_metadata import Integrity, get_data_first_path
from fairspec_metadata import IntegrityType

from fairspec_dataset.actions.file.load import load_file
from fairspec_dataset.actions.file.prefetch import prefetch_files
from fairspec_dataset.actions.stream.concat import concat_file_streams
from fairspec_dataset.actions.stream.load import load_file_stream

if TYPE_CHECKING:
    from fairspec_metadata import Resource


def infer_textual(
    resource: Resource,
    *,
    sample_bytes: int = 10_000,
) -> bool:
    first_path = get_data_first_path(resource)
    if not first_path:
        return False

    buffer = load_file(first_path, max_bytes=sample_bytes)

    if len(buffer) == 0:
        return True

    if _is_binary(buffer):
        return False

    try:
        buffer.decode("utf-8")
        return True
    except UnicodeDecodeError:
        pass

    results = from_bytes(buffer)
    best = results.best()
    if best is not None:
        encoding = best.encoding.lower()
        return encoding in ("utf-8", "ascii")

    return False


def infer_integrity(
    resource: Resource,
    *,
    hash_type: str = "sha256",
) -> Integrity | None:
    hash_value = infer_hash(resource, hash_type=hash_type)

    if not hash_value:
        return None

    return Integrity(type=IntegrityType(hash_type), hash=hash_value)


def infer_hash(
    resource: Resource,
    *,
    hash_type: str = "sha256",
) -> str:
    local_paths = prefetch_files(resource)

    if not local_paths:
        return ""

    streams = [load_file_stream(path) for path in local_paths]
    stream = concat_file_streams(streams)

    h = hashlib.new(hash_type)
    h.update(stream.read())
    return h.hexdigest()


def infer_bytes(resource: Resource) -> int:
    local_paths = prefetch_files(resource)

    total = 0
    for local_path in local_paths:
        total += os.stat(local_path).st_size

    return total


def _is_binary(data: bytes) -> bool:
    control_chars = set(range(0, 8)) | set(range(14, 32))
    return any(byte in control_chars for byte in data)



from __future__ import annotations

import os


def get_is_local_path_exist(path: str) -> bool:
    return os.path.exists(path)


def assert_local_path_vacant(path: str) -> None:
    if os.path.exists(path):
        raise FileExistsError(f'Path "{path}" already exists')



from fairspec_metadata import Integrity, Resource
from fairspec_metadata import IntegrityType

from fairspec_dataset.actions.file.temp import write_temp_file

from .infer import infer_hash
from .validate import validate_file


class TestValidateFile:
    def test_validates_textual_for_utf8(self):
        path = write_temp_file("Hello, World!")
        report = validate_file(Resource(data=path, textual=True))
        assert report.valid is True
        assert report.errors == []

    def test_validates_textual_for_ascii(self):
        path = write_temp_file(b"Simple ASCII text only")
        report = validate_file(Resource(data=path, textual=True))
        assert report.valid is True
        assert report.errors == []

    def test_returns_error_when_textual_expected_but_binary(self):
        path = write_temp_file(bytes([0xFF, 0xD8, 0xFF, 0xE0, 0x00]))
        report = validate_file(Resource(data=path, textual=True))
        assert report.valid is False
        assert len(report.errors) == 1
        assert report.errors[0].type == "file/textual"

    def test_returns_error_when_textual_expected_but_latin1(self):
        buffer = bytes(
            [
                0x43,
                0x61,
                0x66,
                0xE9,
                0x20,
                0x72,
                0xE9,
                0x73,
                0x75,
                0x6D,
                0xE9,
                0x20,
                0x6E,
                0x61,
                0xEF,
                0x76,
                0x65,
                0x20,
                0xE0,
                0x20,
                0x50,
                0x61,
                0x72,
                0x69,
                0x73,
                0x2E,
                0x20,
                0xC7,
                0x61,
                0x20,
                0x63,
                0x27,
                0x65,
                0x73,
                0x74,
                0x20,
                0x62,
                0x6F,
                0x6E,
                0x21,
            ]
        )
        path = write_temp_file(buffer)
        report = validate_file(Resource(data=path, textual=True))
        assert report.valid is False
        assert len(report.errors) == 1
        assert report.errors[0].type == "file/textual"

    def test_validates_integrity_md5(self):
        path = write_temp_file("Hello, World!")
        actual_hash = infer_hash(Resource(data=path), hash_type="md5")
        report = validate_file(
            Resource(
                data=path,
                integrity=Integrity(type=IntegrityType.md5, hash=actual_hash),
            )
        )
        assert report.valid is True
        assert report.errors == []

    def test_returns_error_when_integrity_mismatch(self):
        path = write_temp_file("Hello, World!")
        actual_hash = infer_hash(Resource(data=path), hash_type="md5")
        report = validate_file(
            Resource(
                data=path,
                integrity=Integrity(type=IntegrityType.md5, hash="wronghash"),
            )
        )
        assert report.valid is False
        assert len(report.errors) == 1
        assert report.errors[0].type == "file/integrity"
        assert report.errors[0].hashType == "md5"  # type: ignore[union-attr]
        assert report.errors[0].expectedHash == "wronghash"  # type: ignore[union-attr]
        assert report.errors[0].actualHash == actual_hash  # type: ignore[union-attr]

    def test_validates_sha256_integrity(self):
        path = write_temp_file("Hello, World!")
        actual_hash = infer_hash(Resource(data=path), hash_type="sha256")
        report = validate_file(
            Resource(
                data=path,
                integrity=Integrity(type=IntegrityType.sha256, hash=actual_hash),
            )
        )
        assert report.valid is True
        assert report.errors == []

    def test_validates_sha1_integrity(self):
        path = write_temp_file("Hello, World!")
        actual_hash = infer_hash(Resource(data=path), hash_type="sha1")
        report = validate_file(
            Resource(
                data=path,
                integrity=Integrity(type=IntegrityType.sha1, hash=actual_hash),
            )
        )
        assert report.valid is True
        assert report.errors == []

    def test_validates_sha512_integrity(self):
        path = write_temp_file("Hello, World!")
        actual_hash = infer_hash(Resource(data=path), hash_type="sha512")
        report = validate_file(
            Resource(
                data=path,
                integrity=Integrity(type=IntegrityType.sha512, hash=actual_hash),
            )
        )
        assert report.valid is True
        assert report.errors == []

    def test_validates_both_textual_and_integrity(self):
        path = write_temp_file("Hello, World!")
        actual_hash = infer_hash(Resource(data=path), hash_type="md5")
        report = validate_file(
            Resource(
                data=path,
                textual=True,
                integrity=Integrity(type=IntegrityType.md5, hash=actual_hash),
            )
        )
        assert report.valid is True
        assert report.errors == []

    def test_returns_multiple_errors(self):
        path = write_temp_file(bytes([0xFF, 0xD8, 0xFF, 0xE0, 0x00]))
        report = validate_file(
            Resource(
                data=path,
                textual=True,
                integrity=Integrity(type=IntegrityType.md5, hash="wronghash"),
            )
        )
        assert report.valid is False
        assert len(report.errors) == 2
        assert report.errors[0].type == "file/textual"
        assert report.errors[1].type == "file/integrity"

    def test_returns_error_only_textual_mismatch(self):
        path = write_temp_file(bytes([0xFF, 0xD8, 0xFF, 0xE0, 0x00]))
        actual_hash = infer_hash(Resource(data=path), hash_type="md5")
        report = validate_file(
            Resource(
                data=path,
                textual=True,
                integrity=Integrity(type=IntegrityType.md5, hash=actual_hash),
            )
        )
        assert report.valid is False
        assert len(report.errors) == 1
        assert report.errors[0].type == "file/textual"

    def test_returns_error_only_integrity_mismatch(self):
        path = write_temp_file("Hello, World!")
        report = validate_file(
            Resource(
                data=path,
                textual=True,
                integrity=Integrity(type=IntegrityType.md5, hash="wronghash"),
            )
        )
        assert report.valid is False
        assert len(report.errors) == 1
        assert report.errors[0].type == "file/integrity"

    def test_handles_empty_file(self):
        path = write_temp_file("")
        actual_hash = infer_hash(Resource(data=path), hash_type="sha256")
        report = validate_file(
            Resource(
                data=path,
                integrity=Integrity(type=IntegrityType.sha256, hash=actual_hash),
            )
        )
        assert report.valid is True
        assert report.errors == []

    def test_validates_textual_with_special_characters(self):
        path = write_temp_file("Special: é, ñ, ü, ö, à")
        report = validate_file(Resource(data=path, textual=True))
        assert report.valid is True
        assert report.errors == []



from __future__ import annotations

import json
import os

from fairspec_metadata import ArrayColumnProperty
from fairspec_metadata import BooleanColumnProperty
from fairspec_metadata import DateColumnProperty
from fairspec_metadata import DateTimeColumnProperty
from fairspec_metadata import IntegerColumnProperty
from fairspec_metadata import NumberColumnProperty
from fairspec_metadata import ObjectColumnProperty
from fairspec_metadata import StringColumnProperty
from fairspec_metadata import TimeColumnProperty
from fairspec_metadata import UnknownColumnProperty
from fairspec_metadata import TableSchema

from fairspec_dataset.plugins.ckan.models.schema import CkanSchema
from .from_ckan import convert_table_schema_from_ckan
from .to_ckan import convert_table_schema_to_ckan


def _load_fixture() -> CkanSchema:
    path = os.path.join(os.path.dirname(__file__), "fixtures", "ckan-schema.json")
    with open(path) as f:
        return CkanSchema(**json.load(f))


class TestConvertTableSchemaToCkan:
    def test_converts_fairspec_schema_to_ckan_schema(self):
        schema = TableSchema(
            properties={
                "id": IntegerColumnProperty(
                    type="integer",
                    title="ID",
                    description="Unique identifier",
                ),
                "name": StringColumnProperty(
                    type="string",
                    title="Name",
                    description="Person's full name",
                ),
                "age": IntegerColumnProperty(),
                "score": NumberColumnProperty(
                    type="number",
                    title="Score",
                    description="Test score",
                ),
                "is_active": BooleanColumnProperty(),
                "birth_date": DateColumnProperty(
                    type="string",
                    format="date",
                    title="Birth Date",
                    description="Date of birth",
                ),
                "start_time": TimeColumnProperty(),
                "created_at": DateTimeColumnProperty(
                    type="string",
                    format="date-time",
                    title="Created At",
                    description="Timestamp when record was created",
                ),
                "metadata": ObjectColumnProperty(),
                "tags": ArrayColumnProperty(
                    type="array",
                    title="Tags",
                    description="List of tags",
                ),
            }
        )

        result = convert_table_schema_to_ckan(schema)

        assert schema.properties is not None
        assert result.fields is not None
        assert len(result.fields) == len(schema.properties)

        id_field = next(f for f in result.fields if f.id == "id")
        assert id_field.type == "int"
        assert id_field.info is not None
        assert id_field.info.label == "ID"
        assert id_field.info.notes == "Unique identifier"
        assert id_field.info.type_override == "int"

        name_field = next(f for f in result.fields if f.id == "name")
        assert name_field.type == "text"
        assert name_field.info is not None
        assert name_field.info.label == "Name"
        assert name_field.info.notes == "Person's full name"
        assert name_field.info.type_override == "text"

        age_field = next(f for f in result.fields if f.id == "age")
        assert age_field.type == "int"
        assert age_field.info is None

        score_field = next(f for f in result.fields if f.id == "score")
        assert score_field.type == "numeric"
        assert score_field.info is not None
        assert score_field.info.label == "Score"
        assert score_field.info.notes == "Test score"
        assert score_field.info.type_override == "numeric"

        is_active_field = next(f for f in result.fields if f.id == "is_active")
        assert is_active_field.type == "bool"
        assert is_active_field.info is None

        birth_date_field = next(f for f in result.fields if f.id == "birth_date")
        assert birth_date_field.type == "date"
        assert birth_date_field.info is not None
        assert birth_date_field.info.label == "Birth Date"
        assert birth_date_field.info.notes == "Date of birth"
        assert birth_date_field.info.type_override == "date"

        start_time_field = next(f for f in result.fields if f.id == "start_time")
        assert start_time_field.type == "time"
        assert start_time_field.info is None

        created_at_field = next(f for f in result.fields if f.id == "created_at")
        assert created_at_field.type == "timestamp"
        assert created_at_field.info is not None
        assert created_at_field.info.label == "Created At"
        assert created_at_field.info.notes == "Timestamp when record was created"
        assert created_at_field.info.type_override == "timestamp"

        metadata_field = next(f for f in result.fields if f.id == "metadata")
        assert metadata_field.type == "json"
        assert metadata_field.info is None

        tags_field = next(f for f in result.fields if f.id == "tags")
        assert tags_field.type == "array"
        assert tags_field.info is not None
        assert tags_field.info.label == "Tags"
        assert tags_field.info.notes == "List of tags"
        assert tags_field.info.type_override == "array"

    def test_handles_columns_with_only_title(self):
        schema = TableSchema(
            properties={"field1": StringColumnProperty(title="Field 1")}
        )

        result = convert_table_schema_to_ckan(schema)

        assert result.fields is not None
        assert len(result.fields) == 1
        field = result.fields[0]
        assert field.id == "field1"
        assert field.type == "text"
        assert field.info is not None
        assert field.info.label == "Field 1"
        assert field.info.notes is None
        assert field.info.type_override == "text"

    def test_handles_columns_with_only_description(self):
        schema = TableSchema(
            properties={
                "field1": StringColumnProperty(
                    type="string", description="Field 1 description"
                )
            }
        )

        result = convert_table_schema_to_ckan(schema)

        assert result.fields is not None
        assert len(result.fields) == 1
        field = result.fields[0]
        assert field.id == "field1"
        assert field.type == "text"
        assert field.info is not None
        assert field.info.label is None
        assert field.info.notes == "Field 1 description"
        assert field.info.type_override == "text"

    def test_handles_columns_without_title_or_description(self):
        schema = TableSchema(properties={"simple_field": StringColumnProperty()})

        result = convert_table_schema_to_ckan(schema)

        assert result.fields is not None
        assert len(result.fields) == 1
        field = result.fields[0]
        assert field.id == "simple_field"
        assert field.type == "text"
        assert field.info is None

    def test_handles_empty_properties(self):
        schema = TableSchema(properties={})

        result = convert_table_schema_to_ckan(schema)

        assert result.fields == []

    def test_converts_unmapped_types_to_text(self):
        schema = TableSchema(
            properties={"null_field": UnknownColumnProperty(type="null")}
        )

        result = convert_table_schema_to_ckan(schema)

        assert result.fields is not None
        assert len(result.fields) == 1
        assert result.fields[0].type == "text"

    def test_round_trip_ckan_to_fairspec_to_ckan(self):
        original = _load_fixture()

        fairspec_schema = convert_table_schema_from_ckan(original)
        assert fairspec_schema.properties is not None

        result = convert_table_schema_to_ckan(fairspec_schema)

        assert result.fields is not None
        assert original.fields is not None
        assert len(result.fields) == len(original.fields)

        for original_field in original.fields:
            result_field = next(
                (f for f in result.fields if f.id == original_field.id), None
            )
            assert result_field is not None

            if original_field.info:
                assert result_field.info is not None
                if original_field.info.label:
                    assert result_field.info.label == original_field.info.label
                if original_field.info.notes:
                    assert result_field.info.notes == original_field.info.notes



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from fairspec_metadata import get_is_remote_path

from fairspec_dataset.plugin import DatasetPlugin
from .actions.dataset.load import load_dataset_from_zenodo

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


class ZenodoPlugin(DatasetPlugin):
    def load_dataset(self, source: str) -> Descriptor | None:
        if not _get_is_zenodo(source):
            return None
        return load_dataset_from_zenodo(source)


def _get_is_zenodo(path: str) -> bool:
    if not get_is_remote_path(path):
        return False
    hostname = urllib.parse.urlparse(path).hostname or ""
    return hostname.endswith("zenodo.org")



from .actions.dataset.load import load_dataset_from_zip
from .actions.dataset.save import save_dataset_to_zip
from .plugin import ZipPlugin

__all__ = [
    "ZipPlugin",
    "load_dataset_from_zip",
    "save_dataset_to_zip",
]



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from fairspec_dataset.actions.dataset.merge import merge_datasets

from fairspec_dataset.plugins.ckan.models.dataset import CkanDataset
from fairspec_dataset.plugins.ckan.services.ckan import make_ckan_api_request
from .from_ckan import convert_dataset_from_ckan

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


def load_dataset_from_ckan(
    dataset_url: str,
    *,
    api_key: str | None = None,
) -> Descriptor:
    dataset_id = _extract_dataset_id(dataset_url)
    if not dataset_id:
        raise Exception(f"Failed to extract dataset ID from URL: {dataset_url}")

    ckan_dict = make_ckan_api_request(
        ckan_url=dataset_url,
        action="package_show",
        payload={"id": dataset_id},
        api_key=api_key,
    )

    for resource in ckan_dict.get("resources", []):
        resource_id = resource.get("id")
        if resource.get("format", "").upper() in ("CSV", "XLS", "XLSX"):
            schema = _load_ckan_schema(
                dataset_url=dataset_url,
                resource_id=resource_id,
                api_key=api_key,
            )
            if schema:
                resource["schema"] = schema

    ckan_dataset = CkanDataset(**ckan_dict)
    system_dataset = convert_dataset_from_ckan(ckan_dataset)
    user_dataset_path: str | None = None
    for resource in system_dataset.resources or []:
        custom = resource.unstable_customMetadata or {}
        if custom.get("ckanKey") == "dataset.json":
            user_dataset_path = custom.get("ckanUrl")
            break

    dataset = merge_datasets(
        system_dataset=system_dataset,
        user_dataset_path=user_dataset_path,
    )

    for resource in dataset.resources or []:
        resource.unstable_customMetadata = None

    return dataset.model_dump(by_alias=True, exclude_none=True)


def _extract_dataset_id(dataset_url: str) -> str | None:
    parsed = urllib.parse.urlparse(dataset_url)
    parts = [p for p in parsed.path.split("/") if p]
    return parts[-1] if parts else None


def _load_ckan_schema(
    *,
    dataset_url: str,
    resource_id: str,
    api_key: str | None = None,
) -> dict | None:
    try:
        result = make_ckan_api_request(
            ckan_url=dataset_url,
            action="datastore_search",
            payload={"resource_id": resource_id, "limit": 0},
            api_key=api_key,
        )
        fields = [
            f
            for f in result.get("fields", [])
            if f.get("id") not in ("_id", "_full_text")
        ]
        return {"fields": fields}
    except Exception:
        return None



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from fairspec_dataset.actions.dataset.merge import merge_datasets
from fairspec_dataset.plugins.github.models.repository import GithubRepository

from fairspec_dataset.plugins.github.services.github import make_github_api_request
from .from_github import convert_dataset_from_github

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


def load_dataset_from_github(
    repo_url: str,
    *,
    api_key: str | None = None,
) -> Descriptor:
    owner, repo = _extract_repository_info(repo_url)
    if not owner or not repo:
        raise Exception(f"Failed to extract repository info from URL: {repo_url}")

    repository = make_github_api_request(
        endpoint=f"/repos/{owner}/{repo}",
        api_key=api_key,
    )

    ref = repository["default_branch"]
    tree_response = make_github_api_request(
        endpoint=f"/repos/{owner}/{repo}/git/trees/{ref}?recursive=1",
        api_key=api_key,
    )
    repository["files"] = tree_response["tree"]

    system_dataset = convert_dataset_from_github(GithubRepository(**repository))
    user_dataset_path: str | None = None
    for resource in system_dataset.resources or []:
        custom = resource.unstable_customMetadata or {}
        if custom.get("githubKey") == "dataset.json":
            user_dataset_path = custom.get("githubUrl")
            break

    dataset = merge_datasets(
        system_dataset=system_dataset,
        user_dataset_path=user_dataset_path,
    )

    for resource in dataset.resources or []:
        resource.unstable_customMetadata = None

    return dataset.model_dump(by_alias=True, exclude_none=True)


def _extract_repository_info(repo_url: str) -> tuple[str | None, str | None]:
    parsed = urllib.parse.urlparse(repo_url)
    parts = [p for p in parsed.path.split("/") if p]
    if len(parts) >= 2:
        return parts[0], parts[1]
    return None, None



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import CreatorNameType, DateType, DescriptionType
from fairspec_metadata import Creator, CreatorAffiliation
from fairspec_metadata import DataciteDate
from fairspec_metadata import DataciteDescription
from fairspec_metadata import Rights
from fairspec_metadata import Subject
from fairspec_metadata import Title
from fairspec_metadata import Dataset

from fairspec_dataset.plugins.zenodo.actions.resource.from_zenodo import (
    convert_resource_from_zenodo,
)

if TYPE_CHECKING:
    from fairspec_dataset.plugins.zenodo.models.record import ZenodoRecord


def convert_dataset_from_zenodo(zenodo_record: ZenodoRecord) -> Dataset:
    metadata = zenodo_record.metadata

    titles = [Title(title=metadata.title)] if metadata and metadata.title else None

    descriptions = (
        [
            DataciteDescription(
                description=metadata.description,
                descriptionType=DescriptionType.Abstract,
            )
        ]
        if metadata and metadata.description
        else None
    )

    creators_raw = metadata.creators or [] if metadata else []
    creators = None
    if creators_raw:
        creators = []
        for creator_data in creators_raw:
            affiliation = (
                [CreatorAffiliation(name=creator_data.affiliation)]
                if creator_data.affiliation
                else None
            )
            creators.append(
                Creator(
                    name=creator_data.name or "",
                    nameType=CreatorNameType.Personal,
                    affiliation=affiliation,
                )
            )

    keywords = metadata.keywords or [] if metadata else []
    subjects = [Subject(subject=kw) for kw in keywords] if keywords else None

    dates = (
        [DataciteDate(date=metadata.publication_date, dateType=DateType.Issued)]
        if metadata and metadata.publication_date
        else None
    )

    rights_list = (
        [Rights(rights=metadata.license)] if metadata and metadata.license else None
    )

    doi = metadata.doi if metadata else None
    version = metadata.version if metadata else None

    files = zenodo_record.files or []
    resource_list = [convert_resource_from_zenodo(f) for f in files] if files else []

    return Dataset(
        titles=titles,
        descriptions=descriptions,
        creators=creators,
        subjects=subjects,
        dates=dates,
        rightsList=rights_list,
        doi=doi,
        version=version,
        resources=resource_list,
    )



from unittest.mock import MagicMock, patch

from .plugin import ZenodoPlugin


class TestLoadDataset:
    def setup_method(self):
        self.plugin = ZenodoPlugin()

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_loads_from_zenodo_url(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset("https://zenodo.org/records/123456")
        mock_load.assert_called_once_with("https://zenodo.org/records/123456")
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_returns_none_for_non_zenodo_urls(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://example.com/data")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_returns_none_for_local_paths(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("/tmp/data")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_returns_none_for_github_urls(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("https://github.com/owner/repo")
        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_handles_sandbox_zenodo(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset("https://sandbox.zenodo.org/records/123456")
        mock_load.assert_called_once()
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_handles_zenodo_urls_with_paths(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset("https://zenodo.org/records/123456/files")
        mock_load.assert_called_once()
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_handles_zenodo_urls_with_query_params(self, mock_load: MagicMock):
        mock_load.return_value = {"resources": []}
        result = self.plugin.load_dataset("https://zenodo.org/records/123456?preview=1")
        mock_load.assert_called_once()
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.zenodo.plugin.load_dataset_from_zenodo")
    def test_returns_none_for_http_non_zenodo_urls(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("http://example.com/data")
        mock_load.assert_not_called()
        assert result is None



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from fairspec_metadata import get_file_name_slug
from fairspec_metadata import Integrity, IntegrityType
from fairspec_metadata import Resource

if TYPE_CHECKING:
    from fairspec_dataset.plugins.github.models.file import GithubFile


def convert_resource_from_github(
    github_file: GithubFile,
    *,
    default_branch: str,
) -> Resource:
    path = _convert_path(
        url=github_file.url or "",
        ref=default_branch,
        file_path=github_file.path or "",
    )

    return Resource(
        data=path,
        name=get_file_name_slug(path) or github_file.sha or "",
        integrity=Integrity(
            type=IntegrityType.sha1,
            hash=github_file.sha or "",
        ),
        unstable_customMetadata={
            "githubKey": github_file.path,
            "githubUrl": path,
        },
    )


def _convert_path(*, url: str, ref: str, file_path: str) -> str:
    parsed = urllib.parse.urlparse(url)
    parts = parsed.path.split("/")[2:]
    if len(parts) >= 2:
        owner, repo = parts[0], parts[1]
    else:
        owner, repo = "", ""
    return (
        f"https://raw.githubusercontent.com/{owner}/{repo}/refs/heads/{ref}/{file_path}"
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import (
    ContributorType,
    CreatorNameType,
    DateType,
    DescriptionType,
)
from fairspec_metadata import Contributor
from fairspec_metadata import Creator
from fairspec_metadata import DataciteDate
from fairspec_metadata import DataciteDescription
from fairspec_metadata import Rights
from fairspec_metadata import Subject
from fairspec_metadata import Title
from fairspec_metadata import Dataset

from fairspec_dataset.plugins.ckan.actions.resource.from_ckan import (
    convert_resource_from_ckan,
)

if TYPE_CHECKING:
    from fairspec_dataset.plugins.ckan.models.dataset import CkanDataset


def convert_dataset_from_ckan(ckan_dataset: CkanDataset) -> Dataset:
    titles = [Title(title=ckan_dataset.title)] if ckan_dataset.title else None

    descriptions = (
        [
            DataciteDescription(
                description=ckan_dataset.notes, descriptionType=DescriptionType.Abstract
            )
        ]
        if ckan_dataset.notes
        else None
    )

    version = ckan_dataset.version

    resources = ckan_dataset.resources or []
    resource_list = (
        [convert_resource_from_ckan(r) for r in resources] if resources else []
    )

    rights_list = None
    if ckan_dataset.license_id or ckan_dataset.license_title:
        rights_list = [
            Rights(
                rights=ckan_dataset.license_title or ckan_dataset.license_id or "",
                rightsUri=ckan_dataset.license_url,
                rightsIdentifier=ckan_dataset.license_id,
            )
        ]

    creators = (
        [Creator(name=ckan_dataset.author, nameType=CreatorNameType.Personal)]
        if ckan_dataset.author
        else None
    )

    contributors = None
    if ckan_dataset.maintainer:
        contributors = [
            Contributor(
                name=ckan_dataset.maintainer,
                nameType=CreatorNameType.Personal,
                contributorType=ContributorType.ContactPerson,
            )
        ]

    tags = ckan_dataset.tags or []
    subjects = [Subject(subject=tag.name or "") for tag in tags] if tags else None

    dates: list[DataciteDate] = []
    if ckan_dataset.metadata_created:
        dates.append(
            DataciteDate(date=ckan_dataset.metadata_created, dateType=DateType.Created)
        )
    if ckan_dataset.metadata_modified:
        dates.append(
            DataciteDate(date=ckan_dataset.metadata_modified, dateType=DateType.Updated)
        )

    return Dataset(
        titles=titles,
        descriptions=descriptions,
        version=version,
        resources=resource_list,
        rightsList=rights_list,
        creators=creators,
        contributors=contributors,
        subjects=subjects,
        dates=dates if dates else None,
    )



from __future__ import annotations

from pydantic import ConfigDict, Field

from fairspec_metadata.models.base import FairspecModel

from .schema import CkanSchema


class CkanResource(FairspecModel):
    model_config = ConfigDict(populate_by_name=True)

    id: str | None = None
    url: str | None = None
    name: str | None = None
    created: str | None = None
    description: str | None = None
    format: str | None = None
    hash: str | None = None
    last_modified: str | None = None
    metadata_modified: str | None = None
    mimetype: str | None = None
    size: int | None = None
    schema_: CkanSchema | None = Field(default=None, alias="schema")



from io import BytesIO

from .concat import concat_file_streams


class TestConcatFileStreams:
    def test_concatenates_multiple_streams_in_order(self):
        stream1 = BytesIO(b"Hello, ")
        stream2 = BytesIO(b"World")
        stream3 = BytesIO(b"!")

        result = concat_file_streams([stream1, stream2, stream3])
        assert result.read() == b"Hello, World!"

    def test_handles_single_stream(self):
        stream = BytesIO(b"Single stream content")

        result = concat_file_streams([stream])
        assert result.read() == b"Single stream content"

    def test_handles_empty_array(self):
        result = concat_file_streams([])
        assert result.read() == b""

    def test_handles_streams_with_empty_content(self):
        stream1 = BytesIO(b"")
        stream2 = BytesIO(b"Content")
        stream3 = BytesIO(b"")

        result = concat_file_streams([stream1, stream2, stream3])
        assert result.read() == b"Content"

    def test_concatenates_streams_with_multiple_chunks(self):
        stream1 = BytesIO(b"ABC")
        stream2 = BytesIO(b"DEF")

        result = concat_file_streams([stream1, stream2])
        assert result.read() == b"ABCDEF"

    def test_handles_binary_data(self):
        stream1 = BytesIO(bytes([0x00, 0x01, 0x02]))
        stream2 = BytesIO(bytes([0x03, 0x04, 0x05]))

        result = concat_file_streams([stream1, stream2])
        assert result.read() == bytes([0x00, 0x01, 0x02, 0x03, 0x04, 0x05])

    def test_handles_large_streams(self):
        content1 = b"A" * 5000
        content2 = b"B" * 5000
        stream1 = BytesIO(content1)
        stream2 = BytesIO(content2)

        result = concat_file_streams([stream1, stream2])
        data = result.read()
        assert len(data) == 10000
        assert data == content1 + content2

    def test_preserves_unicode_characters(self):
        stream1 = BytesIO("Hello 世界".encode())
        stream2 = BytesIO(" مرحبا".encode())
        stream3 = BytesIO(" 🌍".encode())

        result = concat_file_streams([stream1, stream2, stream3])
        assert result.read().decode() == "Hello 世界 مرحبا 🌍"

    def test_maintains_stream_order_with_many_streams(self):
        streams = [BytesIO(str(i).encode()) for i in range(10)]

        result = concat_file_streams(streams)
        assert result.read() == b"0123456789"



import os

import pytest
from fairspec_metadata import Resource

from .prefetch import prefetch_files

REMOTE_URL = "https://raw.githubusercontent.com/fairspec/fairspec-typescript/refs/heads/main/table/plugins/csv/actions/table/fixtures/table.csv"


@pytest.mark.vcr
class TestPrefetchFiles:
    def test_prefetches_remote_file(self):
        resource = Resource(data=REMOTE_URL)
        result = prefetch_files(resource)
        assert len(result) == 1
        path = result[0]
        assert os.path.exists(path)
        stats = os.stat(path)
        assert stats.st_size == 27
        with open(path, encoding="utf-8") as f:
            content = f.read()
        assert "id,name" in content

    def test_prefetches_remote_file_with_max_bytes(self):
        resource = Resource(data=REMOTE_URL)
        max_bytes = 18
        result = prefetch_files(resource, max_bytes=max_bytes)
        assert len(result) == 1
        path = result[0]
        assert os.path.exists(path)
        stats = os.stat(path)
        assert stats.st_size == max_bytes
        with open(path, encoding="utf-8") as f:
            content = f.read()
        assert "id,name" in content
        assert "1,english" in content
        assert "中文" not in content



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fairspec_metadata import Descriptor


def convert_resource_to_zenodo(resource: Descriptor) -> dict:
    zenodo_file: dict = {}

    if resource.get("name"):
        zenodo_file["key"] = resource["name"]

    integrity = resource.get("integrity", {})
    if isinstance(integrity, dict) and integrity.get("type") == "md5":
        zenodo_file["checksum"] = f"md5:{integrity['hash']}"

    return zenodo_file



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_dataset.plugins.ckan.actions.resource.to_ckan import (
    convert_resource_to_ckan,
)
from fairspec_dataset.plugins.ckan.models.dataset import CkanDataset
from fairspec_dataset.plugins.ckan.models.tag import CkanTag

if TYPE_CHECKING:
    from fairspec_metadata import Dataset


def convert_dataset_to_ckan(dataset: Dataset) -> CkanDataset:
    title = None
    titles = dataset.titles or []
    if titles and titles[0].title:
        title = titles[0].title

    notes = None
    descriptions = dataset.descriptions or []
    if descriptions and descriptions[0].description:
        notes = descriptions[0].description

    version = dataset.version

    license_id = None
    license_title = None
    license_url = None
    rights_list = dataset.rightsList or []
    if rights_list:
        rights = rights_list[0]
        license_id = rights.rightsIdentifier
        license_title = rights.rights
        license_url = rights.rightsUri

    author = None
    creators = dataset.creators or []
    if creators and creators[0].name:
        author = creators[0].name

    maintainer = None
    contributors = dataset.contributors or []
    if contributors:
        maintainer_obj = next(
            (c for c in contributors if c.contributorType == "ContactPerson"),
            None,
        )
        if maintainer_obj and maintainer_obj.name:
            maintainer = maintainer_obj.name

    resource_list = None
    resources = dataset.resources or []
    if resources:
        resource_list = [
            r
            for r in (convert_resource_to_ckan(res) for res in resources)
            if r is not None
        ]
    else:
        resource_list = []

    tags = None
    subjects = dataset.subjects or []
    if subjects:
        tags = [CkanTag(name=s.subject, display_name=s.subject) for s in subjects]
    else:
        tags = []

    metadata_created = None
    metadata_modified = None
    dates = dataset.dates or []
    created_date = next((d for d in dates if d.dateType == "Created"), None)
    if created_date:
        metadata_created = created_date.date
    updated_date = next((d for d in dates if d.dateType == "Updated"), None)
    if updated_date:
        metadata_modified = updated_date.date

    return CkanDataset(
        title=title,
        notes=notes,
        version=version,
        license_id=license_id,
        license_title=license_title,
        license_url=license_url,
        author=author,
        maintainer=maintainer,
        resources=resource_list,
        tags=tags,
        metadata_created=metadata_created,
        metadata_modified=metadata_modified,
    )



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from fairspec_metadata import denormalize_dataset
from fairspec_metadata import save_descriptor

from fairspec_dataset.actions.dataset.basepath import get_dataset_basepath
from fairspec_dataset.actions.file.copy import copy_file
from fairspec_dataset.actions.file.path import assert_local_path_vacant
from fairspec_dataset.actions.folder.create import create_folder
from fairspec_dataset.actions.resource.save import SaveFileProps, save_resource_files

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor


def save_dataset_to_folder(
    dataset: Dataset,
    *,
    folder_path: str,
    with_remote: bool = False,
) -> Descriptor:
    basepath = get_dataset_basepath(dataset)

    assert_local_path_vacant(folder_path)
    create_folder(folder_path)

    resource_descriptors: list[Descriptor] = []
    for resource in dataset.resources or []:
        resource_descriptors.append(
            save_resource_files(
                resource,
                basepath=basepath,
                with_remote=with_remote,
                save_file=_make_save_file(folder_path),
            )
        )

    denormalized = denormalize_dataset(dataset, basepath=basepath)
    descriptor: Descriptor = {
        **denormalized.model_dump(by_alias=True, exclude_none=True),
        "resources": resource_descriptors,
    }

    save_descriptor(descriptor, path=os.path.join(folder_path, "dataset.json"))

    return descriptor


def _make_save_file(folder_path: str):
    def save_file(props: SaveFileProps) -> str:
        copy_file(
            source_path=props.normalized_path,
            target_path=os.path.join(folder_path, props.denormalized_path),
        )
        return props.denormalized_path

    return save_file



from .actions.dataset.load import load_dataset_from_ckan
from .actions.dataset.save import save_dataset_to_ckan
from .plugin import CkanPlugin

__all__ = [
    "CkanPlugin",
    "load_dataset_from_ckan",
    "save_dataset_to_ckan",
]



from __future__ import annotations

import zipfile
from typing import TYPE_CHECKING

from fairspec_metadata import denormalize_dataset
from fairspec_metadata import stringify_descriptor

from fairspec_dataset.actions.dataset.basepath import get_dataset_basepath
from fairspec_dataset.actions.file.path import assert_local_path_vacant
from fairspec_dataset.actions.resource.save import SaveFileProps, save_resource_files
from fairspec_dataset.actions.stream.load import load_file_stream

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor


def save_dataset_to_zip(
    dataset: Dataset,
    *,
    archive_path: str,
    with_remote: bool = False,
) -> None:
    basepath = get_dataset_basepath(dataset)

    assert_local_path_vacant(archive_path)
    files: dict[str, bytes] = {}

    resource_descriptors: list[Descriptor] = []
    for resource in dataset.resources or []:
        resource_descriptors.append(
            save_resource_files(
                resource,
                basepath=basepath,
                with_remote=with_remote,
                save_file=_make_save_file(files),
            )
        )

    denormalized = denormalize_dataset(dataset, basepath=basepath)
    descriptor: Descriptor = {
        **denormalized.model_dump(by_alias=True, exclude_none=True),
        "resources": resource_descriptors,
    }

    files["dataset.json"] = stringify_descriptor(descriptor).encode("utf-8")

    with zipfile.ZipFile(archive_path, "w", zipfile.ZIP_DEFLATED) as zf:
        for name, data in files.items():
            zf.writestr(name, data)


def _make_save_file(files: dict[str, bytes]):
    def save_file(props: SaveFileProps) -> str:
        stream = load_file_stream(props.normalized_path)
        files[props.denormalized_path] = stream.read()
        return props.denormalized_path

    return save_file



from __future__ import annotations

import os

from fairspec_metadata import load_dataset_descriptor
from fairspec_metadata import Dataset


def load_dataset_from_folder(folder_path: str) -> Dataset:
    return load_dataset_descriptor(os.path.join(folder_path, "dataset.json"))



from __future__ import annotations

from io import BytesIO
from typing import BinaryIO


def concat_file_streams(streams: list[BinaryIO]) -> BinaryIO:
    parts: list[bytes] = []
    for stream in streams:
        parts.append(stream.read())
    return BytesIO(b"".join(parts))



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import get_columns

from fairspec_dataset.plugins.ckan.models.field import CkanField, CkanFieldInfo
from fairspec_dataset.plugins.ckan.models.schema import CkanSchema

if TYPE_CHECKING:
    from fairspec_metadata import Column
    from fairspec_metadata import TableSchema


def convert_table_schema_to_ckan(table_schema: TableSchema) -> CkanSchema:
    fields: list[CkanField] = []

    columns = get_columns(table_schema.model_dump(by_alias=True, exclude_none=True))
    for column in columns:
        fields.append(_convert_column(column))

    return CkanSchema(fields=fields)


def _convert_column(column: Column) -> CkanField:
    title = column.property.title
    description = column.property.description

    info = None
    if title or description:
        info = CkanFieldInfo(
            label=title if title else None,
            notes=description if description else None,
            type_override=_convert_type(column),
        )

    return CkanField(
        id=column.name,
        type=_convert_type(column),
        info=info,
    )


def _convert_type(column: Column) -> str:
    match column.type:
        case "string":
            return "text"
        case "integer":
            return "int"
        case "number":
            return "numeric"
        case "boolean":
            return "bool"
        case "date":
            return "date"
        case "time":
            return "time"
        case "date-time":
            return "timestamp"
        case "object":
            return "json"
        case "array":
            return "array"
        case _:
            return "text"



from __future__ import annotations

import json
import urllib.parse
import urllib.request


def make_zenodo_api_request(
    *,
    endpoint: str,
    method: str = "GET",
    payload: dict | None = None,
    upload: tuple[str, bytes] | None = None,
    api_key: str | None = None,
    sandbox: bool = False,
) -> dict:
    base_url = "https://sandbox.zenodo.org/api" if sandbox else "https://zenodo.org/api"
    url = f"{base_url}{endpoint}"

    if api_key:
        separator = "&" if "?" in url else "?"
        url = f"{url}{separator}access_token={api_key}"

    headers: dict[str, str] = {}
    data: bytes | None = None

    if upload:
        file_name, file_data = upload
        boundary = "----FormBoundary7MA4YWxkTrZu0gW"
        headers["Content-Type"] = f"multipart/form-data; boundary={boundary}"

        body = b""
        body += f"--{boundary}\r\n".encode()
        body += f'Content-Disposition: form-data; name="file"; filename="{file_name}"\r\n'.encode()
        body += b"Content-Type: application/octet-stream\r\n\r\n"
        body += file_data
        body += f"\r\n--{boundary}--\r\n".encode()
        data = body
    elif payload is not None:
        headers["Content-Type"] = "application/json"
        data = json.dumps(payload).encode()

    request = urllib.request.Request(url, data=data, headers=headers, method=method)

    with urllib.request.urlopen(request) as response:
        response_data = response.read().decode()
        if not response_data:
            return {}
        return json.loads(response_data)



from __future__ import annotations

from typing import TYPE_CHECKING

from fairspec_metadata import get_column_properties
from fairspec_metadata import ArrayColumn, ArrayColumnProperty
from fairspec_metadata import BooleanColumn, BooleanColumnProperty
from fairspec_metadata import DateColumn, DateColumnProperty
from fairspec_metadata import (
    DateTimeColumn,
    DateTimeColumnProperty,
)
from fairspec_metadata import IntegerColumn, IntegerColumnProperty
from fairspec_metadata import NumberColumn, NumberColumnProperty
from fairspec_metadata import ObjectColumn, ObjectColumnProperty
from fairspec_metadata import StringColumn, StringColumnProperty
from fairspec_metadata import TimeColumn, TimeColumnProperty
from fairspec_metadata import TableSchema

if TYPE_CHECKING:
    from fairspec_metadata import Column

    from fairspec_dataset.plugins.ckan.models.field import CkanField
    from fairspec_dataset.plugins.ckan.models.schema import CkanSchema


def convert_table_schema_from_ckan(ckan_schema: CkanSchema) -> TableSchema:
    columns: list[Column] = []

    for ckan_field in ckan_schema.fields or []:
        columns.append(_convert_column(ckan_field))

    return TableSchema(properties=get_column_properties(columns))


def _convert_column(ckan_field: CkanField) -> Column:
    info = ckan_field.info

    base_kwargs: dict = {}
    if info:
        if info.label:
            base_kwargs["title"] = info.label
        if info.notes:
            base_kwargs["description"] = info.notes

    column_type = (
        (info.type_override if info else None) or ckan_field.type or "text"
    ).lower()

    name = ckan_field.id or ""

    match column_type:
        case "text" | "string":
            return StringColumn(
                name=name,
                type="string",
                property=StringColumnProperty(type="string", **base_kwargs),
            )
        case "int" | "integer":
            return IntegerColumn(
                name=name,
                type="integer",
                property=IntegerColumnProperty(type="integer", **base_kwargs),
            )
        case "numeric" | "number" | "float":
            return NumberColumn(
                name=name,
                type="number",
                property=NumberColumnProperty(type="number", **base_kwargs),
            )
        case "bool" | "boolean":
            return BooleanColumn(
                name=name,
                type="boolean",
                property=BooleanColumnProperty(type="boolean", **base_kwargs),
            )
        case "date":
            return DateColumn(
                name=name,
                type="date",
                property=DateColumnProperty(
                    type="string", format="date", **base_kwargs
                ),
            )
        case "time":
            return TimeColumn(
                name=name,
                type="time",
                property=TimeColumnProperty(
                    type="string", format="time", **base_kwargs
                ),
            )
        case "timestamp" | "datetime":
            return DateTimeColumn(
                name=name,
                type="date-time",
                property=DateTimeColumnProperty(
                    type="string", format="date-time", **base_kwargs
                ),
            )
        case "json" | "object":
            return ObjectColumn(
                name=name,
                type="object",
                property=ObjectColumnProperty(type="object", **base_kwargs),
            )
        case "array":
            return ArrayColumn(
                name=name,
                type="array",
                property=ArrayColumnProperty(type="array", **base_kwargs),
            )
        case _:
            return StringColumn(
                name=name,
                type="string",
                property=StringColumnProperty(type="string", **base_kwargs),
            )



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Unpack

from fairspec_metadata import get_is_remote_path

from fairspec_dataset.models.dataset import SaveDatasetResult
from fairspec_dataset.plugin import DatasetPlugin
from .actions.dataset.load import load_dataset_from_folder
from .actions.dataset.save import save_dataset_to_folder

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor

    from fairspec_dataset.models.dataset import SaveDatasetOptions


class FolderPlugin(DatasetPlugin):
    def load_dataset(self, source: str) -> Descriptor | None:
        if not _get_is_folder(source):
            return None
        dataset = load_dataset_from_folder(source)
        return dataset.model_dump(by_alias=True, exclude_none=True)

    def save_dataset(
        self, dataset: Dataset, **options: Unpack[SaveDatasetOptions]
    ) -> SaveDatasetResult | None:
        target = options["target"]
        if not _get_is_folder(target):
            return None
        save_dataset_to_folder(
            dataset, folder_path=target, with_remote=bool(options.get("with_remote"))
        )
        return SaveDatasetResult(path=target)


def _get_is_folder(path: str) -> bool:
    if get_is_remote_path(path):
        return False
    try:
        return os.path.isdir(path)
    except Exception:
        return False



from __future__ import annotations

from typing import TYPE_CHECKING, Unpack

from fairspec_metadata import get_file_extension

from fairspec_dataset.models.dataset import SaveDatasetResult
from fairspec_dataset.plugin import DatasetPlugin
from .actions.dataset.load import load_dataset_from_zip
from .actions.dataset.save import save_dataset_to_zip

if TYPE_CHECKING:
    from fairspec_metadata import Dataset
    from fairspec_metadata import Descriptor

    from fairspec_dataset.models.dataset import SaveDatasetOptions


class ZipPlugin(DatasetPlugin):
    def load_dataset(self, source: str) -> Descriptor | None:
        if not _get_is_zip(source):
            return None
        dataset = load_dataset_from_zip(source)
        return dataset.model_dump(by_alias=True, exclude_none=True)

    def save_dataset(
        self, dataset: Dataset, **options: Unpack[SaveDatasetOptions]
    ) -> SaveDatasetResult | None:
        target = options["target"]
        if not _get_is_zip(target):
            return None
        save_dataset_to_zip(
            dataset, archive_path=target, with_remote=bool(options.get("with_remote"))
        )
        return SaveDatasetResult(path=None)


def _get_is_zip(path: str) -> bool:
    return get_file_extension(path) == "zip"



from unittest.mock import MagicMock, patch

from fairspec_metadata import Dataset
from fairspec_metadata import Resource

from .plugin import ZipPlugin


class TestLoadDataset:
    def setup_method(self):
        self.plugin = ZipPlugin()

    @patch("fairspec_dataset.plugins.zip.plugin.load_dataset_from_zip")
    def test_loads_from_zip_file(self, mock_load: MagicMock):
        mock_dataset = MagicMock()
        mock_dataset.model_dump.return_value = {
            "resources": [{"name": "test-resource", "data": []}]
        }
        mock_load.return_value = mock_dataset

        result = self.plugin.load_dataset("test.zip")

        mock_load.assert_called_once_with("test.zip")
        assert result == {"resources": [{"name": "test-resource", "data": []}]}

    @patch("fairspec_dataset.plugins.zip.plugin.load_dataset_from_zip")
    def test_returns_none_for_non_zip(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("test.json")

        mock_load.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.zip.plugin.load_dataset_from_zip")
    def test_handles_paths_with_directories(self, mock_load: MagicMock):
        mock_dataset = MagicMock()
        mock_dataset.model_dump.return_value = {"resources": []}
        mock_load.return_value = mock_dataset

        result = self.plugin.load_dataset("/path/to/file.zip")

        mock_load.assert_called_once_with("/path/to/file.zip")
        assert result == {"resources": []}

    @patch("fairspec_dataset.plugins.zip.plugin.load_dataset_from_zip")
    def test_returns_none_for_no_extension(self, mock_load: MagicMock):
        result = self.plugin.load_dataset("test")

        mock_load.assert_not_called()
        assert result is None


class TestSaveDataset:
    def setup_method(self):
        self.plugin = ZipPlugin()
        self.dataset = Dataset(resources=[Resource(name="test_resource", data=[])])

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_saves_to_zip_file(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="output.zip")

        mock_save.assert_called_once_with(
            self.dataset, archive_path="output.zip", with_remote=False
        )
        assert result is not None
        assert result.path is None

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_returns_none_for_non_zip(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="output.json")

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_passes_with_remote_option(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(
            self.dataset, target="output.zip", with_remote=True
        )

        mock_save.assert_called_once_with(
            self.dataset, archive_path="output.zip", with_remote=True
        )
        assert result is not None

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_with_remote_defaults_to_false(self, mock_save: MagicMock):
        self.plugin.save_dataset(self.dataset, target="output.zip")

        mock_save.assert_called_once_with(
            self.dataset, archive_path="output.zip", with_remote=False
        )

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_handles_paths_with_directories(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="/path/to/output.zip")

        mock_save.assert_called_once_with(
            self.dataset, archive_path="/path/to/output.zip", with_remote=False
        )
        assert result is not None

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_returns_none_for_no_extension(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="output")

        mock_save.assert_not_called()
        assert result is None

    @patch("fairspec_dataset.plugins.zip.plugin.save_dataset_to_zip")
    def test_result_has_path_none(self, mock_save: MagicMock):
        result = self.plugin.save_dataset(self.dataset, target="output.zip")

        assert result is not None
        assert result.path is None
